# Copyright (c) OpenMMLab. All rights reserved.
import os
import random

import numpy as np
import torch
import torch.distributed as dist
from antgo.framework.helper.runner import (DistSamplerSeedHook, EpochBasedRunner,
                         Fp16OptimizerHook, OptimizerHook, build_optimizer,
                         build_runner, get_dist_info)
from antgo.framework.helper.dataset import (build_dataloader, build_dataset)
from antgo.framework.helper.utils.compat_config import *
from antgo.framework.helper.utils import get_logger
from antgo.framework.helper.utils import find_latest_checkpoint
from antgo.framework.helper.runner.hooks.evaluation import *
from antgo.framework.helper.utils.util_distribution import build_ddp, build_dp, get_device
from antgo.framework.helper.runner.builder import *

def init_random_seed(seed=None, device='cuda'):
    """Initialize random seed.

    If the seed is not set, the seed will be automatically randomized,
    and then broadcast to all processes to prevent some potential bugs.

    Args:
        seed (int, Optional): The seed. Default to None.
        device (str): The device where the seed will be put on.
            Default to 'cuda'.

    Returns:
        int: Seed to be used.
    """
    if seed is not None:
        return seed

    # Make sure all ranks share the same random seed to prevent
    # some potential bugs. Please refer to
    # https://github.com/open-mmlab/mmdetection/issues/6339
    rank, world_size = get_dist_info()
    seed = np.random.randint(2**31)
    if world_size == 1:
        return seed

    if rank == 0:
        random_num = torch.tensor(seed, dtype=torch.int32, device=device)
    else:
        random_num = torch.tensor(0, dtype=torch.int32, device=device)
    dist.broadcast(random_num, src=0)
    return random_num.item()


def set_random_seed(seed, deterministic=False):
    """Set random seed.

    Args:
        seed (int): Seed to be used.
        deterministic (bool): Whether to set the deterministic option for
            CUDNN backend, i.e., set `torch.backends.cudnn.deterministic`
            to True and `torch.backends.cudnn.benchmark` to False.
            Default: False.
    """
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    if deterministic:
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False


def auto_scale_lr(cfg, distributed, logger):
    """Automatically scaling LR according to GPU number and sample per GPU.

    Args:
        cfg (config): Training config.
        distributed (bool): Using distributed or not.
        logger (logging.Logger): Logger.
    """
    # Get flag from config
    if ('auto_scale_lr' not in cfg) or \
            (not cfg.auto_scale_lr.get('enable', False)):
        logger.info('Automatic scaling of learning rate (LR)'
                    ' has been disabled.')
        return

    # Get base batch size from config
    base_batch_size = cfg.auto_scale_lr.get('base_batch_size', None)
    if base_batch_size is None:
        return

    # Get gpu number
    if distributed:
        _, world_size = get_dist_info()
        num_gpus = len(range(world_size))
    else:
        num_gpus = len(cfg.gpu_ids)

    # calculate the batch size
    samples_per_gpu = cfg.data.train_dataloader.samples_per_gpu
    batch_size = num_gpus * samples_per_gpu
    logger.info(f'Training with {num_gpus} GPU(s) with {samples_per_gpu} '
                f'samples per GPU. The total batch size is {batch_size}.')

    if batch_size != base_batch_size:
        # scale LR with
        # [linear scaling rule](https://arxiv.org/abs/1706.02677)
        scaled_lr = (batch_size / base_batch_size) * cfg.optimizer.lr
        logger.info('LR has been automatically scaled '
                    f'from {cfg.optimizer.lr} to {scaled_lr}')
        cfg.optimizer.lr = scaled_lr
    else:
        logger.info('The batch size match the '
                    f'base batch size: {base_batch_size}, '
                    f'will not scaling the LR ({cfg.optimizer.lr}).')


def train_model(model,
                   dataset,
                   cfg,
                   distributed=False,
                   validate=False,
                   timestamp=None,
                   meta=None):
    logger = get_logger('model', log_level=cfg.log_level)

    # prepare data loaders
    dataset = dataset if isinstance(dataset, (list, tuple)) else [dataset]

    runner_type = 'EpochBasedRunner' if 'runner' not in cfg else cfg.runner['type']

    train_dataloader_default_args = dict(
        samples_per_gpu=2,
        workers_per_gpu=2,
        # `num_gpus` will be ignored if distributed
        num_gpus=len(cfg.gpu_ids),
        dist=distributed,
        seed=cfg.seed,
        runner_type=runner_type,
        persistent_workers=False)

    train_loader_cfg = {
        **train_dataloader_default_args,
        **cfg.data.get('train_dataloader', {})
    }
    data_loaders = [build_dataloader(ds, **train_loader_cfg) for ds in dataset]

    # put model on gpus
    if distributed:
        find_unused_parameters = cfg.get('find_unused_parameters', False)
        # Sets the `find_unused_parameters` parameter in
        # torch.nn.parallel.DistributedDataParallel
        model = build_ddp(
            model,
            cfg.device,
            device_ids=[int(os.environ['LOCAL_RANK'])],
            broadcast_buffers=False,
            find_unused_parameters=find_unused_parameters)
    else:
        model = build_dp(model, cfg.device, device_ids=cfg.gpu_ids)

    # build optimizer
    auto_scale_lr(cfg, distributed, logger)
    optimizer = build_optimizer(model, cfg.optimizer)

    # build training strategy
    runner = build_runner(
        cfg.runner,
        default_args=dict(
            model=model,
            optimizer=optimizer,
            work_dir=cfg.work_dir,
            logger=logger,
            meta=meta))

    # an ugly workaround to make .log and .log.json filenames the same
    runner.timestamp = timestamp

    # fp16 setting
    fp16_cfg = cfg.get('fp16', None)
    if fp16_cfg is not None:
        optimizer_config = Fp16OptimizerHook(
            **cfg.optimizer_config, **fp16_cfg, distributed=distributed)
    elif distributed and 'type' not in cfg.optimizer_config:
        optimizer_config = OptimizerHook(**cfg.optimizer_config)
    else:
        optimizer_config = cfg.optimizer_config

    # register hooks
    runner.register_training_hooks(
        cfg.lr_config,                                          # 学习率调整策略，比如step,warmup等
        optimizer_config,                                       # 优化器的相关后处理，比如限制梯度操作等
        cfg.checkpoint_config,                                  # checkpoint相关处理
        cfg.log_config,                                         
        cfg.get('momentum_config', None),                       
        custom_hooks_config=cfg.get('custom_hooks', None))

    if distributed:
        if isinstance(runner, EpochBasedRunner):
            runner.register_hook(DistSamplerSeedHook())

    # register eval hooks
    if validate:
        val_dataloader_default_args = dict(
            samples_per_gpu=1,
            workers_per_gpu=2,
            dist=distributed,
            shuffle=False,
            persistent_workers=False)

        val_dataloader_args = {
            **val_dataloader_default_args,
            **cfg.data.get('val_dataloader', {})
        }
        val_dataset = build_dataset(cfg.data.val, dict(test_mode=True))
        val_dataloader = build_dataloader(val_dataset, **val_dataloader_args)
        eval_cfg = cfg.get('evaluation', {})
        eval_cfg['by_epoch'] = cfg.runner['type'] != 'IterBasedRunner'
        eval_hook = DistEvalHook if distributed else EvalHook

        if 'metric' in eval_cfg:
            metric = build_measure(eval_cfg['metric'])
            eval_cfg['metric'] = metric
        runner.register_hook(
            eval_hook(val_dataloader, **eval_cfg), priority='LOW')

    resume_from = None
    if cfg.resume_from is None and cfg.get('auto_resume'):
        resume_from = find_latest_checkpoint(cfg.work_dir)
    if resume_from is not None:
        cfg.resume_from = resume_from

    if cfg.resume_from:
        runner.resume(cfg.resume_from)
    elif cfg.load_from:
        runner.load_checkpoint(cfg.load_from)
    runner.run(data_loaders, cfg.workflow, mode='train')



import cv2
import numpy as np
import torch
from pytorch3d.transforms.rotation_conversions import matrix_to_axis_angle


def get_rot6s(pose):
    num_pose = pose.shape[0]
    rot6s = list()
    for i in range(num_pose):
        rot6 = cv2.Rodrigues(pose[i])[0][:2]
        rot6s.append(rot6.reshape(-1))
    rot6s = np.array(rot6s)
    return rot6s


def matrix2dof(matrixs):
    # batch_size = matrixs.shape[0]
    # dofs = list()
    # for i in range(batch_size):
    #     dof = cv2.Rodrigues(matrixs[i].numpy())
    #     dofs.append(dof[0])
    # dofs = np.array(dofs)
    dofs = matrix_to_axis_angle(matrixs).reshape(-1, 3, 1)
    return dofs


def rot6tomatrix(rot6s):
    poses = rot6s.reshape(-1, 6)
    x_raw = poses[:, 0:3]  # batch*3
    y_raw = poses[:, 3:6]  # batch*3

    x = normalize_vector(x_raw)  # batch*3
    y = normalize_vector(y_raw)  # batch*3
    middle = normalize_vector(x + y)
    orthmid = normalize_vector(x - y)
    x = normalize_vector(middle + orthmid)
    y = normalize_vector(middle - orthmid)
    # Their scalar product should be small !
    # assert torch.einsum("ij,ij->i", [x, y]).abs().max() < 0.00001
    z = normalize_vector(cross_product(x, y))

    x = x.view(-1, 3, 1)
    y = y.view(-1, 3, 1)
    z = z.view(-1, 3, 1)
    matrix = torch.cat((x, y, z), 2)  # batch*3*3
    # Check for reflection in matrix ! If found, flip last vector TODO

    # torch.det too slow, skip
    # assert (torch.stack([torch.det(mat) for mat in matrix]) < 0).sum() == 0

    return matrix


def normalize_vector(v):
    batch = v.shape[0]
    v_mag = torch.sqrt(v.pow(2).sum(1))  # batch
    v_mag = torch.max(v_mag, v.new([1e-8]))
    v_mag = v_mag.view(batch, 1).expand(batch, v.shape[1])
    v = v / v_mag
    return v


def cross_product(u, v):
    batch = u.shape[0]
    i = u[:, 1] * v[:, 2] - u[:, 2] * v[:, 1]
    j = u[:, 2] * v[:, 0] - u[:, 0] * v[:, 2]
    k = u[:, 0] * v[:, 1] - u[:, 1] * v[:, 0]

    out = torch.cat((i.view(batch, 1), j.view(batch, 1), k.view(batch, 1)), 1)

    return out


def rot6todof(rot6s):
    # rot6s = torch.from_numpy(rot6s)
    matrix = rot6tomatrix(rot6s)
    output_dofs = matrix2dof(matrix)
    output_dofs = output_dofs.reshape(-1, 3)

    return output_dofs * -1


def eular_to_mat(roll, pitch, yaw):
    rx = np.asarray(
        [
            [1.0, 0.0, 0.0],
            [0.0, np.cos(roll), -np.sin(roll)],
            [0.0, np.sin(roll), np.cos(roll)],
        ]
    )
    ry = np.asarray(
        [
            [np.cos(pitch), 0.0, np.sin(pitch)],
            [0.0, 1.0, 0.0],
            [-np.sin(pitch), 0.0, np.cos(pitch)],
        ]
    )
    rz = np.asarray(
        [
            [np.cos(yaw), -np.sin(yaw), 0.0],
            [np.sin(yaw), np.cos(yaw), 0.0],
            [0.0, 0.0, 1.0],
        ]
    )
    return rx @ ry @ rz



from antgo.framework.helper.parallel import is_module_wrapper
from antgo.framework.helper.runner.hooks import HOOKS, Hook
from bisect import bisect_right


@HOOKS.register_module()
class Weighter(Hook):
    def __init__(
        self,
        steps=None,
        vals=None,
        name=None,
    ):
        self.steps = steps
        self.vals = vals
        self.name = name
        if self.name is not None:
            assert self.steps is not None
            assert self.vals is not None
            assert len(self.vals) == len(self.steps) + 1

    def before_train_iter(self, runner):
        curr_step = runner.iter
        if self.name is None:
            return
        model = runner.model
        if is_module_wrapper(model):
            model = model.module
        assert hasattr(model, self.name)
        self.steps = [s if s > 0 else runner.max_iters - s for s in self.steps]
        runner.log_buffer.output[self.name] = self.vals[
            bisect_right(self.steps, curr_step)
        ]

        setattr(model, self.name, runner.log_buffer.output[self.name])



import os
from antgo.pipeline import extent
from antgo.pipeline.extent.glue.common import *
import copy

 
class CppOp(object):
    def __init__(self, *args, func_op_name, **kwargs):
        # get cpp func
        self.func_op_name = func_op_name
        self.func = getattr(extent.func, func_op_name)

        # parameter
        self.args = list(args) + [None] * (len(self.func.func.arg_names)-len(args))
        for i, (var_name, var_type) in enumerate(zip(self.func.func.arg_names, self.func.func.arg_types)):
            if kwargs.get(var_name, None) is not None and var_type.is_const:
                self.args[i] = kwargs.get(var_name)

    def __call__(self, *args):
        # 输入数据部分
        func_args = copy.deepcopy(self.args)

        # 绑定上游数据 (顺序绑定)
        input_i = 0
        for i, var_type in enumerate(self.func.func.arg_types):
            if var_type.is_const and func_args[i] is None and input_i < len(args):
                func_args[i] = args[input_i]
                input_i += 1

        # 输出数据(仅用于占位)
        type_map = {
            'CDTensor': np.float64,
            'CFTensor': np.float32,
            'CITensor': np.int32,
            'CUCTensor': np.uint8
        }
        out_placeholders = [np.empty((1), type_map[self.func.func.arg_types[i].cname.replace('*', '')]) \
                for i in range(len(self.func.func.arg_types)) if not self.func.func.arg_types[i].is_const]

        # 绑定输出数据（顺序绑定）
        output_i = 0
        for i, var_type in enumerate(self.func.func.arg_types):
            if not var_type.is_const and func_args[i] is None and output_i < len(args):
                func_args[i] = out_placeholders[output_i]
                output_i += 1

        # 检查函数参数
        for func_arg in func_args:
            if func_arg is None:
                print(f'cpp func {self.func_op_name} arg abnormal.')
                print(func_args)
                break

        # 运行
        output_data = self.func(*func_args)

        # 返回结果
        return tuple(output_data) if len(output_data) > 1 else output_data[0]



from numbers import Number

import numpy as np
import torch
import torch.nn as nn


def accuracy_numpy(pred, target, topk=(1, ), thrs=0.):
    if isinstance(thrs, Number):
        thrs = (thrs, )
        res_single = True
    elif isinstance(thrs, tuple):
        res_single = False
    else:
        raise TypeError(
            f'thrs should be a number or tuple, but got {type(thrs)}.')

    res = []
    maxk = max(topk)
    num = pred.shape[0]

    static_inds = np.indices((num, maxk))[0]
    pred_label = pred.argpartition(-maxk, axis=1)[:, -maxk:]
    pred_score = pred[static_inds, pred_label]

    sort_inds = np.argsort(pred_score, axis=1)[:, ::-1]
    pred_label = pred_label[static_inds, sort_inds]
    pred_score = pred_score[static_inds, sort_inds]

    for k in topk:
        correct_k = pred_label[:, :k] == target.reshape(-1, 1)
        res_thr = []
        for thr in thrs:
            # Only prediction values larger than thr are counted as correct
            _correct_k = correct_k & (pred_score[:, :k] > thr)
            _correct_k = np.logical_or.reduce(_correct_k, axis=1)
            res_thr.append((_correct_k.sum() * 100. / num))
        if res_single:
            res.append(res_thr[0])
        else:
            res.append(res_thr)
    return res


def accuracy_torch(pred, target, topk=(1, ), thrs=0.):
    if isinstance(thrs, Number):
        thrs = (thrs, )
        res_single = True
    elif isinstance(thrs, tuple):
        res_single = False
    else:
        raise TypeError(
            f'thrs should be a number or tuple, but got {type(thrs)}.')

    res = []
    maxk = max(topk)
    num = pred.size(0)
    pred = pred.float()
    pred_score, pred_label = pred.topk(maxk, dim=1)
    pred_label = pred_label.t()
    correct = pred_label.eq(target.view(1, -1).expand_as(pred_label))
    for k in topk:
        res_thr = []
        for thr in thrs:
            # Only prediction values larger than thr are counted as correct
            _correct = correct & (pred_score.t() > thr)
            correct_k = _correct[:k].reshape(-1).float().sum(0, keepdim=True)
            res_thr.append((correct_k.mul_(100. / num)))
        if res_single:
            res.append(res_thr[0])
        else:
            res.append(res_thr)
    return res


def accuracy(pred, target, topk=1, thrs=0.):
    """Calculate accuracy according to the prediction and target.

    Args:
        pred (torch.Tensor | np.array): The model prediction.
        target (torch.Tensor | np.array): The target of each prediction
        topk (int | tuple[int]): If the predictions in ``topk``
            matches the target, the predictions will be regarded as
            correct ones. Defaults to 1.
        thrs (Number | tuple[Number], optional): Predictions with scores under
            the thresholds are considered negative. Default to 0.

    Returns:
        torch.Tensor | list[torch.Tensor] | list[list[torch.Tensor]]: Accuracy
            - torch.Tensor: If both ``topk`` and ``thrs`` is a single value.
            - list[torch.Tensor]: If one of ``topk`` or ``thrs`` is a tuple.
            - list[list[torch.Tensor]]: If both ``topk`` and ``thrs`` is a \
              tuple. And the first dim is ``topk``, the second dim is ``thrs``.
    """
    assert isinstance(topk, (int, tuple))
    if isinstance(topk, int):
        topk = (topk, )
        return_single = True
    else:
        return_single = False

    assert isinstance(pred, (torch.Tensor, np.ndarray)), \
        f'The pred should be torch.Tensor or np.ndarray ' \
        f'instead of {type(pred)}.'
    assert isinstance(target, (torch.Tensor, np.ndarray)), \
        f'The target should be torch.Tensor or np.ndarray ' \
        f'instead of {type(target)}.'

    # torch version is faster in most situations.
    to_tensor = (lambda x: torch.from_numpy(x)
                 if isinstance(x, np.ndarray) else x)
    pred = to_tensor(pred)
    target = to_tensor(target)

    res = accuracy_torch(pred, target, topk, thrs)

    return res[0] if return_single else res


class Accuracy(nn.Module):

    def __init__(self, topk=(1, )):
        """Module to calculate the accuracy.

        Args:
            topk (tuple): The criterion used to calculate the
                accuracy. Defaults to (1,).
        """
        super().__init__()
        self.topk = topk

    def forward(self, pred, target):
        """Forward function to calculate accuracy.

        Args:
            pred (torch.Tensor): Prediction of models.
            target (torch.Tensor): Target for each prediction.

        Returns:
            list[torch.Tensor]: The accuracies under different topk criterions.
        """
        return accuracy(pred, target, self.topk)



from ..builder import BBOX_CODERS
from ..transforms import bbox2distance, distance2bbox
from .base_bbox_coder import BaseBBoxCoder


@BBOX_CODERS.register_module()
class DistancePointBBoxCoder(BaseBBoxCoder):
    """Distance Point BBox coder.

    This coder encodes gt bboxes (x1, y1, x2, y2) into (top, bottom, left,
    right) and decode it back to the original.

    Args:
        clip_border (bool, optional): Whether clip the objects outside the
            border of the image. Defaults to True.
    """

    def __init__(self, clip_border=True):
        super(BaseBBoxCoder, self).__init__()
        self.clip_border = clip_border

    def encode(self, points, gt_bboxes, max_dis=None, eps=0.1):
        """Encode bounding box to distances.

        Args:
            points (Tensor): Shape (N, 2), The format is [x, y].
            gt_bboxes (Tensor): Shape (N, 4), The format is "xyxy"
            max_dis (float): Upper bound of the distance. Default None.
            eps (float): a small value to ensure target < max_dis, instead <=.
                Default 0.1.

        Returns:
            Tensor: Box transformation deltas. The shape is (N, 4).
        """
        assert points.size(0) == gt_bboxes.size(0)
        assert points.size(-1) == 2
        assert gt_bboxes.size(-1) == 4
        return bbox2distance(points, gt_bboxes, max_dis, eps)

    def decode(self, points, pred_bboxes, max_shape=None):
        """Decode distance prediction to bounding box.

        Args:
            points (Tensor): Shape (B, N, 2) or (N, 2).
            pred_bboxes (Tensor): Distance from the given point to 4
                boundaries (left, top, right, bottom). Shape (B, N, 4)
                or (N, 4)
            max_shape (Sequence[int] or torch.Tensor or Sequence[
                Sequence[int]],optional): Maximum bounds for boxes, specifies
                (H, W, C) or (H, W). If priors shape is (B, N, 4), then
                the max_shape should be a Sequence[Sequence[int]],
                and the length of max_shape should also be B.
                Default None.
        Returns:
            Tensor: Boxes with shape (N, 4) or (B, N, 4)
        """
        assert points.size(0) == pred_bboxes.size(0)
        assert points.size(-1) == 2
        assert pred_bboxes.size(-1) == 4
        if self.clip_border is False:
            max_shape = None
        return distance2bbox(points, pred_bboxes, max_shape)



# Copyright (c) OpenMMLab. All rights reserved.
from .brick_wrappers import AdaptiveAvgPool2d, adaptive_avg_pool2d
from .builder import build_linear_layer, build_transformer
from .ckpt_convert import pvt_convert
from .conv_upsample import ConvUpsample
from .csp_layer import CSPLayer
from .gaussian_target import gaussian_radius, gen_gaussian_target
from .inverted_residual import InvertedResidual
from .make_divisible import make_divisible
from .misc import interpolate_as, sigmoid_geometric_mean
from .normed_predictor import NormedConv2d, NormedLinear
from .panoptic_gt_processing import preprocess_panoptic_gt
from .res_layer import ResLayer, SimplifiedBasicBlock
from .se_layer import DyReLU, SELayer

__all__ = [
    'ResLayer', 'gaussian_radius', 'gen_gaussian_target',
    'DetrTransformerDecoderLayer', 'DetrTransformerDecoder', 'Transformer',
    'build_transformer', 'build_linear_layer', 'DynamicConv', 'SimplifiedBasicBlock',
    'NormedLinear', 'NormedConv2d', 'make_divisible', 'InvertedResidual',
    'SELayer', 'interpolate_as', 'ConvUpsample', 'CSPLayer',
    'adaptive_avg_pool2d', 'AdaptiveAvgPool2d', 'PatchEmbed', 'nchw_to_nlc',
    'nlc_to_nchw', 'pvt_convert', 'sigmoid_geometric_mean',
    'preprocess_panoptic_gt', 'DyReLU'
]



# -*- coding: UTF-8 -*-
# @Time    : 2022/9/21 22:42
# @File    : inference_model_op.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

from antgo.pipeline.engine import *
from antgo.pipeline.models.utils.utils import *
from antgo.framework.helper.utils.config import Config
from antgo.framework.helper.models.builder import *
from antgo.framework.helper.runner.checkpoint import load_checkpoint
from antgo.framework.helper.utils.util_distribution import build_ddp, build_dp, get_device
from antgo.framework.helper.utils import build_from_cfg
from antgo.framework.helper.parallel.collate import collate
from antgo.framework.helper.cnn.utils import fuse_conv_bn
import torch
import cv2
import json
import copy
import numpy as np


@register
class inference_model_op(object):
  def __init__(self, cfg_file=None, checkpoint='', gpu_id=-1, is_fuse_conv_bn=False, output=None):
    self.device = 'cpu' if gpu_id < 0 else 'cuda'
    self.cfg = Config.fromfile(cfg_file)

    # build model
    self.model = build_model(self.cfg.model)
    if checkpoint == '':
      checkpoint = self.cfg.get('checkpoint', checkpoint)
    checkpoint = load_checkpoint(self.model, checkpoint, map_location='cpu')

    if is_fuse_conv_bn:
      print('use fuse conv_bn')
      self.model = fuse_conv_bn(self.model)
    self.model = build_dp(self.model, self.device, device_ids=[gpu_id])

    # inputs_def = self.cfg.data.test.inputs_def
    # self._fields = copy.deepcopy(inputs_def['fields']) if inputs_def else None
    self._fields = ['image', 'image_meta']
    self._output = output

  def _arrange(self, sample, fields):
    if type(fields[0]) == list or type(fields[0]) == tuple:
      warp_ins = []
      for field in fields:
        one_ins = {}
        for ff in field:
          one_ins[ff] = sample[ff]

        warp_ins.append(one_ins)
      return warp_ins

    warp_ins = {}
    for field in fields:
      warp_ins[field] = sample[field]

    return warp_ins

  def __call__(self, image):
    # 输入是组batch前的最后一步数据 (C,H,W)
    sample = {
      'image': torch.from_numpy(image),
      'image_meta': {
        'image_shape': image.shape[1:],
        'scale_factor': (1, 1, 1, 1),
        'transform_matrix': np.eye(3),
      }
    }

    # arange warp
    sample = self._arrange(sample, self._fields)
    sample = collate([sample])
    sample.update({
      'return_loss': False
    })

    # model run
    self.model.eval()
    result = {}
    with torch.no_grad():
      result = self.model(**sample)

    if self._output is None:
      self._output = result.keys()
    out = []
    for k in self._output:
      out.append(result[k][0].detach().cpu().numpy())
    return tuple(out)







# -*- coding: UTF-8 -*-
# @Time    : 18-3-26
# @File    : fashionai_attribute.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import os
import sys
import numpy as np
from antgo.dataflow.dataset import *

__all__ = ['FashionAIAttribute']
class FashionAIAttribute(Dataset):
  def __init__(self, train_or_test, dir=None, params=None):
    super(FashionAIAttribute, self).__init__(train_or_test, dir, params)
    assert(train_or_test in ['train', 'test'])
    self.train_or_test = train_or_test
    
    if self.train_or_test == 'train':
      if not os.path.exists(os.path.join(self.dir, 'base')) or not os.path.exists(os.path.join(self.dir, 'show')):
        logger.error('FashionAIAttribute must be download from \n https://tianchi.aliyun.com/competition/information.htm?spm=5176.100071.5678.2.778c2b9eBime0R&raceId=231649')
        sys.exit(0)
    else:
      if not os.path.exists(os.path.join(self.dir , 'rank')):
        logger.error('FashionAIAttribute must be download from \n https://tianchi.aliyun.com/competition/information.htm?spm=5176.100071.5678.2.778c2b9eBime0R&raceId=231649')
        sys.exit(0)
    
    self.images = []
    self.cloth_attribs = {
      'skirt_length_labels': ['Invisible', 'Short Length', 'Knee Length', 'Midi Length', 'Ankle Length', 'Floor Length'],
      'coat_length_labels': ['Invisible', 'High Waist Length', 'Regular Length', 'Long Length', 'Micro Length', 'Knee Length', 'Midi Length', 'Ankle&Floor Length'],
      'collar_design_labels': ['Invisible', 'Shirt Collar', 'Peter Pan', 'Puritan Collar', 'Rib Collar'],
      'lapel_design_labels': ['Invisible', 'Notched', 'Collarless', 'Shawl Collar', 'Plus Size Shawl'],
      'neck_design_labels': ['Invisible', 'Turtle Neck', 'Ruffle Semi-High Collar', 'Low Turtle Neck', 'Draped Collar'],
      'neckline_design_labels': ['Invisible', 'Strapless Neck', 'Deep V Neckline', 'Straight Neck', 'V Neckline', 'Square Neckline', 'Off Shoulder', 'Round Neckline', 'Sweat Heart Neck', 'OneShoulder Neckline'],
      'pant_length_labels': ['Invisible', 'Short Pant', 'Mid Length', '3/4 Length', 'Cropped Pant', 'Full Length'],
      'sleeve_length_labels': ['Invisible', 'Sleeveless', 'Cup Sleeves', 'Short Sleeves', 'Elbow Sleeves', '3/4 Sleeves','Wrist Length','Long Sleeves','Extra Long Sleeves']
    }
    
    self.cloth_attribs_index = {
      'skirt_length_labels': 0,
      'coat_length_labels': 1,
      'collar_design_labels': 2,
      'lapel_design_labels': 3,
      'neck_design_labels': 4,
      'neckline_design_labels': 5,
      'pant_length_labels': 6,
      'sleeve_length_labels': 7,
    }
    
    if self.train_or_test == 'train':
      # 1.step parse from base folder
      with open(os.path.join(self.dir, 'base',  'Annotations', 'label.csv')) as fp:
        content = fp.readline()
        while content:
          image_file, cloth_attrib, cloth_attrib_value = content.split(',')
          cloth_attrib_value = cloth_attrib_value.replace('\n','')
          assert(len(self.cloth_attribs[cloth_attrib]) == len(cloth_attrib_value))
          cloth_attrib_label = np.zeros((len(cloth_attrib_value)), dtype=np.int32)
          for i in range(len(cloth_attrib_value)):
            if cloth_attrib_value[i] == 'y':
              cloth_attrib_label[i] = 1
            elif cloth_attrib_value[i] == 'n':
              cloth_attrib_label[i] = 0
            else:
              cloth_attrib_label[i] = -1
              
          self.images.append((image_file, cloth_attrib, cloth_attrib_label, 'base'))
          content = fp.readline()
      # 2.step parse from show folder
      with open(os.path.join(self.dir, 'show', 'Annotations', 'skirt_length_labels.csv')) as fp:
        content = fp.readline()
        while content:
          image_file, cloth_attrib, cloth_attrib_value = content.split(',')
          cloth_attrib_value = cloth_attrib_value.replace('\n', '')
          assert (len(self.cloth_attribs[cloth_attrib]) == len(cloth_attrib_value))
          cloth_attrib_label = np.zeros((len(cloth_attrib_value)), dtype=np.int32)
          for i in range(len(cloth_attrib_value)):
            if cloth_attrib_value[i] == 'y':
              cloth_attrib_label[i] = 1
            elif cloth_attrib_value[i] == 'n':
              cloth_attrib_label[i] = 0
            else:
              cloth_attrib_label[i] = -1
    
          self.images.append((image_file, cloth_attrib, cloth_attrib_label, 'show'))
          content = fp.readline()
    else:
      with open(os.path.join(self.dir, 'rank', 'Tests', 'question.csv')) as fp:
        content = fp.readline()
        while content:
          image_file, cloth_attrib, _ = content.split(',')
          self.images.append((image_file, cloth_attrib, None, 'rank'))
          content = fp.readline()
    
    # data index list
    self.ids = list(range(len(self.images)))
    # fixed seed
    self.seed = getattr(self, 'seed', 0)
  
  @property
  def size(self):
    return len(self.ids)
  
  def data_pool(self):
    epoch = 0
    while True:
      max_epoches = self.epochs if self.epochs is not None else 1
      if epoch >= max_epoches:
        break
      epoch += 1
    
      idxs = copy.deepcopy(self.ids)
      if self.rng:
        self.rng.shuffle(idxs)
    
      for k in idxs:
        image_file, cloth_attrib, cloth_attrib_label, sub_folder = self.images[k]
        image_path = os.path.join(self.dir, sub_folder, image_file)
        image = imread(image_path)
      
        if self.train_or_test == 'train':
          yield [(image, self.cloth_attribs_index[cloth_attrib], image_file),
                 {'category_id': cloth_attrib_label, 'category': cloth_attrib, 'id': k}]
        else:
          yield [(image, self.cloth_attribs_index[cloth_attrib], image_file), None]
    
  def at(self, id):
    image_file, cloth_attrib, cloth_attrib_label, sub_folder = self.images[id]
    image_path = os.path.join(self.dir, sub_folder, image_file)
    image = imread(image_path)
    
    return (image, cloth_attrib, image_file)
  
  def split(self, split_params={}, split_method='holdout'):
    assert(self.train_or_test == 'train')
    assert (split_method in ['repeated-holdout', 'bootstrap', 'kfold', 'holdout'])

    # set fixed random seed
    np.random.seed(np.int64(self.seed))
    category_ids = None
    if split_method == 'kfold':
      category_ids = [i for i in range(len(self.ids))]
      np.random.shuffle(category_ids)
    else:
      category_ids = [0 for _ in range(len(self.ids))]

    train_ids, val_ids = self._split(category_ids, split_params, split_method)
    train_dataset = FashionAIAttribute(self.train_or_test, self.dir, self.ext_params)
    train_dataset.ids = train_ids

    val_dataset = FashionAIAttribute(self.train_or_test, self.dir, self.ext_params)
    val_dataset.ids = val_ids

    return train_dataset, val_dataset  # split data by their label



from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.task.task import *
from antgo.measures.base import *
import numpy as np

class AntTopK(AntMeasure):
  def __init__(self, task):
    super(AntTopK, self).__init__(task, 'ACCURACY')
    assert(task.task_type == 'CLASSIFICATION')
    self.is_support_rank = True
    self.larger_is_better = 1
    self.topk = (1,5)

def eva(self, data, label):
    if label is not None:
      data = zip(data, label)

    maxk = max(self.topk)
    res = [0 for _ in range(len(self.topk))]
    sample_scores = []
    for predict, gt in data:
        # predict: probability
        # gt: {'category_id': label, 'id': ..}
        id = gt['id']
        gt_label = gt['category_id']
        top_idx = predict.argsort()[-maxk:][::-1]

        correct = top_idx == gt_label
        sample_score = []
        for i, k in enumerate(self.topk):
            correct_k = np.sum(correct[:k].reshape(-1).astype(np.float32))
            sample_score[i] = correct_k
            res[i] += correct_k

        sample_scores.append({'id': id, 'score': sample_score[0], 'category': gt_label})
    
    totle_num = len(sample_scores)
    res[0] = res[0] / totle_num
    res[1] = res[1] / totle_num

    return {
      'statistic': {
        'name': self.name,
        'value': [
          {'name': 'top_1', 'value': float(res[0]), 'type': 'SCALAR'},
          {'name': 'top_5', 'value': float(res[1]), 'type': 'SCALAR'},
        ]
      },
      'info': sample_scores
    }


# Copyright (c) OpenMMLab. All rights reserved.
import warnings

import torch
import torch.nn.functional as F

from ..builder import BBOX_ASSIGNERS
from ..iou_calculators import bbox_overlaps
from .assign_result import AssignResult
from .base_assigner import BaseAssigner

@BBOX_ASSIGNERS.register_module()
class SimOTAAssigner(BaseAssigner):
    """Computes matching between predictions and ground truth.

    Args:
        center_radius (int | float, optional): Ground truth center size
            to judge whether a prior is in center. Default 2.5.
        candidate_topk (int, optional): The candidate top-k which used to
            get top-k ious to calculate dynamic-k. Default 10.
        iou_weight (int | float, optional): The scale factor for regression
            iou cost. Default 3.0.
        cls_weight (int | float, optional): The scale factor for classification
            cost. Default 1.0.
        num_classes (int):  Number of categories. Default 80 classes for 
            COCO dataset.
        use_vfl (bool): Whether to use varifocal loss to calculate classification 
            cost. Default False.
    """

    def __init__(self,
                 center_radius=2.5,
                 candidate_topk=10,
                 iou_weight=3.0,
                 cls_weight=1.0,
                 num_classes=80,
                 use_vfl=False):
        self.center_radius = center_radius
        self.candidate_topk = candidate_topk
        self.iou_weight = iou_weight
        self.cls_weight = cls_weight
        self.num_classes = num_classes
        self.use_vfl = use_vfl

    def assign(self,
               pred_scores,
               priors,
               decoded_bboxes,
               gt_bboxes,
               gt_labels,
               gt_bboxes_ignore=None,
               eps=1e-7):
        """Assign gt to priors using SimOTA. It will switch to CPU mode when
        GPU is out of memory.
        Args:
            pred_scores (Tensor): Classification scores of one image,
                a 2D-Tensor with shape [num_priors, num_classes]
            priors (Tensor): All priors of one image, a 2D-Tensor with shape
                [num_priors, 4] in [cx, xy, stride_w, stride_y] format.
            decoded_bboxes (Tensor): Predicted bboxes, a 2D-Tensor with shape
                [num_priors, 4] in [tl_x, tl_y, br_x, br_y] format.
            gt_bboxes (Tensor): Ground truth bboxes of one image, a 2D-Tensor
                with shape [num_gts, 4] in [tl_x, tl_y, br_x, br_y] format.
            gt_labels (Tensor): Ground truth labels of one image, a Tensor
                with shape [num_gts].
            gt_bboxes_ignore (Tensor, optional): Ground truth bboxes that are
                labelled as `ignored`, e.g., crowd boxes in COCO.
            eps (float): A value added to the denominator for numerical
                stability. Default 1e-7.
        Returns:
            assign_result (obj:`AssignResult`): The assigned result.
        """
        try:
            assign_result = self._assign(pred_scores, priors, decoded_bboxes,
                                         gt_bboxes, gt_labels,
                                         gt_bboxes_ignore, eps)
            return assign_result
        except RuntimeError:
            origin_device = pred_scores.device
            warnings.warn('OOM RuntimeError is raised due to the huge memory '
                          'cost during label assignment. CPU mode is applied '
                          'in this batch. If you want to avoid this issue, '
                          'try to reduce the batch size or image size.')
            torch.cuda.empty_cache()

            pred_scores = pred_scores.cpu()
            priors = priors.cpu()
            decoded_bboxes = decoded_bboxes.cpu()
            gt_bboxes = gt_bboxes.cpu().float()
            gt_labels = gt_labels.cpu()

            assign_result = self._assign(pred_scores, priors, decoded_bboxes,
                                         gt_bboxes, gt_labels,
                                         gt_bboxes_ignore, eps)
            assign_result.gt_inds = assign_result.gt_inds.to(origin_device)
            assign_result.max_overlaps = assign_result.max_overlaps.to(
                origin_device)
            assign_result.labels = assign_result.labels.to(origin_device)

            return assign_result

    def _assign(self,
                pred_scores,
                priors,
                decoded_bboxes,
                gt_bboxes,
                gt_labels,
                gt_bboxes_ignore=None,
                eps=1e-7):
        """Assign gt to priors using SimOTA.
        Args:
            pred_scores (Tensor): Classification scores of one image,
                a 2D-Tensor with shape [num_priors, num_classes]
            priors (Tensor): All priors of one image, a 2D-Tensor with shape
                [num_priors, 4] in [cx, xy, stride_w, stride_y] format.
            decoded_bboxes (Tensor): Predicted bboxes, a 2D-Tensor with shape
                [num_priors, 4] in [tl_x, tl_y, br_x, br_y] format.
            gt_bboxes (Tensor): Ground truth bboxes of one image, a 2D-Tensor
                with shape [num_gts, 4] in [tl_x, tl_y, br_x, br_y] format.
            gt_labels (Tensor): Ground truth labels of one image, a Tensor
                with shape [num_gts].
            gt_bboxes_ignore (Tensor, optional): Ground truth bboxes that are
                labelled as `ignored`, e.g., crowd boxes in COCO.
            eps (float): A value added to the denominator for numerical
                stability. Default 1e-7.
        Returns:
            :obj:`AssignResult`: The assigned result.
        """
        INF = 100000000
        num_gt = gt_bboxes.size(0)
        num_bboxes = decoded_bboxes.size(0)

        # assign 0 by default
        assigned_gt_inds = decoded_bboxes.new_full((num_bboxes, ),
                                                   0,
                                                   dtype=torch.long)
        valid_mask, is_in_boxes_and_center = self.get_in_gt_and_in_center_info(
            priors, gt_bboxes)
        valid_decoded_bbox = decoded_bboxes[valid_mask]
        valid_pred_scores = pred_scores[valid_mask]
        num_valid = valid_decoded_bbox.size(0)

        if num_gt == 0 or num_bboxes == 0 or num_valid == 0:
            # No ground truth or boxes, return empty assignment
            max_overlaps = decoded_bboxes.new_zeros((num_bboxes, ))
            if num_gt == 0:
                # No truth, assign everything to background
                assigned_gt_inds[:] = 0
            if gt_labels is None:
                assigned_labels = None
            else:
                assigned_labels = decoded_bboxes.new_full((num_bboxes, ),
                                                          -1,
                                                          dtype=torch.long)
            return AssignResult(
                num_gt, assigned_gt_inds, max_overlaps, labels=assigned_labels)

        pairwise_ious = bbox_overlaps(valid_decoded_bbox, gt_bboxes)
        
        if self.use_vfl:
            gt_vfl_labels = gt_labels.squeeze(-1).unsqueeze(0).repeat(num_valid, 1).reshape(-1)
            valid_pred_scores = valid_pred_scores.unsqueeze(1).repeat(1, num_gt, 1).reshape(-1, self.num_classes)
            vfl_score = pairwise_ious.new_zeros(valid_pred_scores.shape)
            vfl_score[torch.arange(0, vfl_score.shape[0]), gt_vfl_labels] = pairwise_ious.reshape(-1)
            losses_vfl = varifocal_loss(
                valid_pred_scores, vfl_score, use_sigmoid=False).reshape(num_valid, num_gt)
            losses_giou = 1 - bbox_overlaps(
                valid_decoded_bbox, gt_bboxes, mode='giou')
            cost_matrix = (
                losses_vfl * self.cls_weight + losses_giou * self.iou_weight +
                (~is_in_boxes_and_center) * INF)
        else:
            iou_cost = -torch.log(pairwise_ious + eps)

            gt_onehot_label = (
                F.one_hot(gt_labels.to(torch.int64),
                        pred_scores.shape[-1]).float().unsqueeze(0).repeat(
                            num_valid, 1, 1))

            valid_pred_scores = valid_pred_scores.unsqueeze(1).repeat(1, num_gt, 1)
            cls_cost = F.binary_cross_entropy(
                valid_pred_scores.sqrt_(), gt_onehot_label,
                reduction='none').sum(-1)

            cost_matrix = (
                cls_cost * self.cls_weight + iou_cost * self.iou_weight +
                (~is_in_boxes_and_center) * INF)

        matched_pred_ious, matched_gt_inds = \
            self.dynamic_k_matching(
                cost_matrix, pairwise_ious, num_gt, valid_mask)

        # convert to AssignResult format
        assigned_gt_inds[valid_mask] = matched_gt_inds + 1
        assigned_labels = assigned_gt_inds.new_full((num_bboxes, ), -1)
        assigned_labels[valid_mask] = gt_labels[matched_gt_inds].long()
        max_overlaps = assigned_gt_inds.new_full((num_bboxes, ),
                                                 -INF,
                                                 dtype=torch.float32)
        max_overlaps[valid_mask] = matched_pred_ious
        return AssignResult(
            num_gt, assigned_gt_inds, max_overlaps, labels=assigned_labels)

    def get_in_gt_and_in_center_info(self, priors, gt_bboxes):
        num_gt = gt_bboxes.size(0)

        repeated_x = priors[:, 0].unsqueeze(1).repeat(1, num_gt)
        repeated_y = priors[:, 1].unsqueeze(1).repeat(1, num_gt)
        repeated_stride_x = priors[:, 2].unsqueeze(1).repeat(1, num_gt)
        repeated_stride_y = priors[:, 3].unsqueeze(1).repeat(1, num_gt)

        # is prior centers in gt bboxes, shape: [n_prior, n_gt]
        l_ = repeated_x - gt_bboxes[:, 0]
        t_ = repeated_y - gt_bboxes[:, 1]
        r_ = gt_bboxes[:, 2] - repeated_x
        b_ = gt_bboxes[:, 3] - repeated_y

        deltas = torch.stack([l_, t_, r_, b_], dim=1)
        is_in_gts = deltas.min(dim=1).values > 0
        is_in_gts_all = is_in_gts.sum(dim=1) > 0

        # is prior centers in gt centers
        gt_cxs = (gt_bboxes[:, 0] + gt_bboxes[:, 2]) / 2.0
        gt_cys = (gt_bboxes[:, 1] + gt_bboxes[:, 3]) / 2.0
        ct_box_l = gt_cxs - self.center_radius * repeated_stride_x
        ct_box_t = gt_cys - self.center_radius * repeated_stride_y
        ct_box_r = gt_cxs + self.center_radius * repeated_stride_x
        ct_box_b = gt_cys + self.center_radius * repeated_stride_y

        cl_ = repeated_x - ct_box_l
        ct_ = repeated_y - ct_box_t
        cr_ = ct_box_r - repeated_x
        cb_ = ct_box_b - repeated_y

        ct_deltas = torch.stack([cl_, ct_, cr_, cb_], dim=1)
        is_in_cts = ct_deltas.min(dim=1).values > 0
        is_in_cts_all = is_in_cts.sum(dim=1) > 0

        # in boxes or in centers, shape: [num_priors]
        is_in_gts_or_centers = is_in_gts_all | is_in_cts_all

        # both in boxes and centers, shape: [num_fg, num_gt]
        is_in_boxes_and_centers = (
            is_in_gts[is_in_gts_or_centers, :]
            & is_in_cts[is_in_gts_or_centers, :])
        return is_in_gts_or_centers, is_in_boxes_and_centers

    def dynamic_k_matching(self, cost, pairwise_ious, num_gt, valid_mask):
        matching_matrix = torch.zeros_like(cost)
        # select candidate topk ious for dynamic-k calculation
        candidate_topk = min(self.candidate_topk, pairwise_ious.size(0))
        topk_ious, _ = torch.topk(pairwise_ious, candidate_topk, dim=0)
        # calculate dynamic k for each gt
        dynamic_ks = torch.clamp(topk_ious.sum(0).int(), min=1)
        for gt_idx in range(num_gt):
            _, pos_idx = torch.topk(
                cost[:, gt_idx], k=dynamic_ks[gt_idx].item(), largest=False)
            matching_matrix[:, gt_idx][pos_idx] = 1.0

        del topk_ious, dynamic_ks, pos_idx

        prior_match_gt_mask = matching_matrix.sum(1) > 1
        if prior_match_gt_mask.sum() > 0:
            cost_min, cost_argmin = torch.min(
                cost[prior_match_gt_mask, :], dim=1)
            matching_matrix[prior_match_gt_mask, :] *= 0.0
            matching_matrix[prior_match_gt_mask, cost_argmin] = 1.0
        # get foreground mask inside box and center prior
        fg_mask_inboxes = matching_matrix.sum(1) > 0.0
        valid_mask[valid_mask.clone()] = fg_mask_inboxes

        matched_gt_inds = matching_matrix[fg_mask_inboxes, :].argmax(1)
        matched_pred_ious = (matching_matrix *
                             pairwise_ious).sum(1)[fg_mask_inboxes]
        return matched_pred_ious, matched_gt_inds

def varifocal_loss(pred,
                   target,
                   weight=None,
                   alpha=0.75,
                   gamma=2.0,
                   use_sigmoid=True,
                   iou_weighted=True,
                   reduction='mean',
                   avg_factor=None):
    """`Varifocal Loss <https://arxiv.org/abs/2008.13367>`_

    Args:
        pred (torch.Tensor): The prediction with shape (N, C), C is the
            number of classes
        target (torch.Tensor): The learning target of the iou-aware
            classification score with shape (N, C), C is the number of classes.
        weight (torch.Tensor, optional): The weight of loss for each
            prediction. Defaults to None.
        alpha (float, optional): A balance factor for the negative part of
            Varifocal Loss, which is different from the alpha of Focal Loss.
            Defaults to 0.75.
        gamma (float, optional): The gamma for calculating the modulating
            factor. Defaults to 2.0.
        iou_weighted (bool, optional): Whether to weight the loss of the
            positive example with the iou target. Defaults to True.
        reduction (str, optional): The method used to reduce the loss into
            a scalar. Defaults to 'mean'. Options are "none", "mean" and
            "sum".
        avg_factor (int, optional): Average factor that is used to average
            the loss. Defaults to None.
    """
    # pred and target should be of the same size
    assert pred.size() == target.size()
    if use_sigmoid:
        pred_ = pred.sigmoid()
    else:
        pred_ = pred
    target = target.type_as(pred_)
    if iou_weighted:
        focal_weight = target * (target > 0.0).float() + \
            alpha * (pred_ - target).abs().pow(gamma) * \
            (target <= 0.0).float()
    else:
        focal_weight = (target > 0.0).float() + \
            alpha * (pred_ - target).abs().pow(gamma) * \
            (target <= 0.0).float()
    if use_sigmoid:
        loss = F.binary_cross_entropy_with_logits(
            pred, target, reduction='none') * focal_weight
        loss = weight_reduce_loss(loss, weight, reduction, avg_factor)
    else:
        loss = F.binary_cross_entropy(
            pred, target, reduction='none') * focal_weight
        loss = loss.sum(axis=1)
    return loss


# Copyright (c) OpenMMLab. All rights reserved.
import os
import random
import sys
import time
import warnings
from getpass import getuser
from socket import gethostname

import numpy as np
import torch
from .dist_utils import  get_dist_info

def get_host_info():
    """Get hostname and username.

    Return empty string if exception raised, e.g. ``getpass.getuser()`` will
    lead to error in docker container
    """
    host = ''
    try:
        host = f'{getuser()}@{gethostname()}'
    except Exception as e:
        warnings.warn(f'Host or user not found: {str(e)}')
    finally:
        return host


def get_time_str():
    return time.strftime('%Y-%m-%dx%H-%M-%S', time.localtime())


def obj_from_dict(info, parent=None, default_args=None):
    """Initialize an object from dict.

    The dict must contain the key "type", which indicates the object type, it
    can be either a string or type, such as "list" or ``list``. Remaining
    fields are treated as the arguments for constructing the object.

    Args:
        info (dict): Object types and arguments.
        parent (:class:`module`): Module which may containing expected object
            classes.
        default_args (dict, optional): Default arguments for initializing the
            object.

    Returns:
        any type: Object built from the dict.
    """
    assert isinstance(info, dict) and 'type' in info
    assert isinstance(default_args, dict) or default_args is None
    args = info.copy()
    obj_type = args.pop('type')
    if mmcv.is_str(obj_type):
        if parent is not None:
            obj_type = getattr(parent, obj_type)
        else:
            obj_type = sys.modules[obj_type]
    elif not isinstance(obj_type, type):
        raise TypeError('type must be a str or valid type, but '
                        f'got {type(obj_type)}')
    if default_args is not None:
        for name, value in default_args.items():
            args.setdefault(name, value)
    return obj_type(**args)


def set_random_seed(seed, deterministic=False, use_rank_shift=False):
    """Set random seed.

    Args:
        seed (int): Seed to be used.
        deterministic (bool): Whether to set the deterministic option for
            CUDNN backend, i.e., set `torch.backends.cudnn.deterministic`
            to True and `torch.backends.cudnn.benchmark` to False.
            Default: False.
        rank_shift (bool): Whether to add rank number to the random seed to
            have different random seed in different threads. Default: False.
    """
    if use_rank_shift:
        rank, _ = mmcv.runner.get_dist_info()
        seed += rank
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    os.environ['PYTHONHASHSEED'] = str(seed)
    if deterministic:
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False


def init_random_seed(seed=None, device='cuda'):
    """Initialize random seed.

    If the seed is not set, the seed will be automatically randomized,
    and then broadcast to all processes to prevent some potential bugs.

    Args:
        seed (int, Optional): The seed. Default to None.
        device (str): The device where the seed will be put on.
            Default to 'cuda'.

    Returns:
        int: Seed to be used.
    """
    if seed is not None:
        return seed

    # Make sure all ranks share the same random seed to prevent
    # some potential bugs. Please refer to
    # https://github.com/open-mmlab/mmdetection/issues/6339
    rank, world_size = get_dist_info()
    seed = np.random.randint(2**31)
    if world_size == 1:
        return seed

    if rank == 0:
        random_num = torch.tensor(seed, dtype=torch.int32, device=device)
    else:
        random_num = torch.tensor(0, dtype=torch.int32, device=device)
    dist.broadcast(random_num, src=0)
    return random_num.item()


def auto_scale_lr(cfg, distributed, logger):
    """Automatically scaling LR according to GPU number and sample per GPU.

    Args:
        cfg (config): Training config.
        distributed (bool): Using distributed or not.
        logger (logging.Logger): Logger.
    """
    # Get flag from config
    if ('auto_scale_lr' not in cfg) or \
            (not cfg.auto_scale_lr.get('enable', False)):
        logger.info('Automatic scaling of learning rate (LR)'
                    ' has been disabled.')
        return

    # Get base batch size from config
    base_batch_size = cfg.auto_scale_lr.get('base_batch_size', None)
    if base_batch_size is None:
        return

    # Get gpu number
    if distributed:
        _, world_size = get_dist_info()
        num_gpus = len(range(world_size))
    else:
        num_gpus = len(cfg.gpu_ids)

    # calculate the batch size
    samples_per_gpu = cfg.data.train_dataloader.samples_per_gpu
    batch_size = num_gpus * samples_per_gpu
    logger.info(f'Training with {num_gpus} GPU(s) with {samples_per_gpu} '
                f'samples per GPU. The total batch size is {batch_size}.')

    if batch_size != base_batch_size:
        # scale LR with
        # [linear scaling rule](https://arxiv.org/abs/1706.02677)
        scaled_lr = (batch_size / base_batch_size) * cfg.optimizer.lr
        logger.info('LR has been automatically scaled '
                    f'from {cfg.optimizer.lr} to {scaled_lr}')
        cfg.optimizer.lr = scaled_lr
    else:
        logger.info('The batch size match the '
                    f'base batch size: {base_batch_size}, '
                    f'will not scaling the LR ({cfg.optimizer.lr}).')




# Copyright (c) 2021  PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import paddle
import copy
import warnings

TYPE_ORDER = ["bool", "int32", "int64", "float16", "float32", "float64"]
TYPE_MAPPER = {"fp16": "float16", "fp32": "float32", "fp64": "float64"}

def abs(input, *, out=None):
    return paddle.abs(input)


def add(input, other, *, out=None):
    return paddle.add(input, other)


def arange(start,
           end,
           step=1,
           *,
           out=None,
           dtype=None,
           layout=None,
           device=None,
           requires_grad=False):
    if requires_grad:
        return paddle.arange(start, end, step, dtype).requires_grad_(True)
    else:
        return paddle.arange(start, end, step, dtype)


def clip(input, min, max, *, out=None):
    return paddle.clip(input, min, max)


def concat(tensors, dim=0):
    x = tensors
    last_index = -1
    for ele in x:
        t = str(ele.dtype).lower().strip().split(".")[-1]
        if t in TYPE_MAPPER:
            t = TYPE_MAPPER[t]
        index = TYPE_ORDER.index(t)
        if last_index < index:
            last_index = index
    real_type = TYPE_ORDER[last_index]
    x = list(x)
    for i in range(len(x)):
        x[i] = x[i].cast(real_type)
    return paddle.concat(x, dim)


def create_tensor(*size):
    if len(size) > 1:
        return paddle.zeros(size, dtype="float32")
    else:
        return paddle.to_tensor(size[0])


def create_float32_tensor(*size):
    if len(size) > 1:
        return paddle.zeros(size, dtype="float32")
    else:
        out = paddle.to_tensor(size[0])
        out = paddle.cast(out, "float32")
        return out


def create_uint8_tensor(*size):
    if len(size) > 1:
        return paddle.zeros(size, dtype="uint8")
    else:
        out = paddle.to_tensor(size[0])
        out = paddle.cast(out, "uint8")
        return out


def exp(input, *, out=None):
    return paddle.exp(input)


def full(size,
         fill_value,
         *,
         out=None,
         dtype=None,
         layout=None,
         device=None,
         requires_grad=False):
    if requires_grad:
        return paddle.full(size, fill_value, dtype).requires_grad_(True)
    else:
        return paddle.full(size, fill_value, dtype)


def full_like(input,
              fill_value,
              *,
              dtype=None,
              layout=None,
              device=None,
              requires_grad=False,
              memory_format=None):
    if requires_grad:
        return paddle.full_like(input, fill_value, dtype).requires_grad_(True)
    else:
        return paddle.full_like(input, fill_value, dtype)


def linspace(start,
             end,
             steps,
             out=None,
             dtype=None,
             layout=None,
             device=None,
             requires_grad=False):
    if requires_grad:
        return paddle.linspace(start, end, step, dtype).requires_grad_(True)
    else:
        return paddle.linspace(start, end, step, dtype)


def load(f, map_location=None, pickle_module=None, **pickle_load_args):
    return paddle.load(f)


def load_state_dict_from_url(url,
                             model_dir=None,
                             map_location=None,
                             progress=True,
                             check_hash=False,
                             file_name=None):
    return paddle.load(url)


def load_url(url,
             model_dir=None,
             map_location=None,
             progress=True,
             check_hash=False,
             file_name=None):
    return paddle.load(url)


def log(input, *, out=None):
    return paddle.log(input)


def logical_and(input, other, *, out=None):
    return paddle.logical_and(input, other, out)


def logical_not(input, *, out=None):
    return paddle.logical_not(input, out)


def logical_or(input, other, *, out=None):
    return paddle.logical_or(input, other, out)


def logical_xor(input, other, *, out=None):
    return paddle.logical_xor(input, other, out)


def matmul(input, other, *, out=None):
    return paddle.matmul(input, other)


def mul(input, other, *, out=None):
    return paddle.multiply(input, other)


def max(input, dim_other=None, keepdim=False, *, out=None):
    if dim_other is None:
        return paddle.max(input)
    elif isinstance(dim_other, paddle.Tensor):
        return paddle.maximum(input, dim_other)
    else:
        return paddle.max(input, axis=dim_other, keepdim=keepdim)


def mean(input, dim=None, keepdim=False, *, out=None):
    if dim is None:
        warnings.warn('The output of paddle.mean is not scalar!')
        return paddle.mean(input)
    else:
        return paddle.mean(input, axis=dim, keepdim=keepdim)


def min(input, dim_other=None, keepdim=False, *, out=None):
    if dim_other is None:
        return paddle.min(input)
    elif isinstance(dim_other, paddle.Tensor):
        return paddle.minimum(input, dim_other)
    else:
        return paddle.min(input, axis=dim_other, keepdim=keepdim)


def ones(*size,
         out=None,
         dtype=None,
         layout=None,
         device=None,
         requires_grad=False):
    if len(size) == 1 and isinstance(size[0], (tuple, list)):
        shape = size[0]
    else:
        shape = size
    if requires_grad:
        return paddle.ones(shape, dtype).requires_grad_(True)
    else:
        return paddle.ones(shape, dtype)


def ones_like(input,
              *,
              dtype=None,
              layout=None,
              device=None,
              requires_grad=False,
              memory_format=None):
    if requires_grad:
        return paddle.ones_like(input, dtype).requires_grad_(True)
    else:
        return paddle.ones_like(input, dtype)


def set_cuda_device(device):
    if isinstance(device, int):
        return paddle.set_device("gpu:{}".format(device))
    else:
        return paddle.set_device("gpu")


def rand(*size,
         out=None,
         dtype=None,
         layout=None,
         device=None,
         requires_grad=False):
    if len(size) == 1 and isinstance(size[0], (tuple, list)):
        shape = size[0]
    else:
        shape = size
    if requires_grad:
        return paddle.rand(shape, dtype).requires_grad_(True)
    else:
        return paddle.rand(shape, dtype)


def randn(*size,
          out=None,
          dtype=None,
          layout=None,
          device=None,
          requires_grad=False):
    if len(size) == 1 and isinstance(size[0], (tuple, list)):
        shape = size[0]
    else:
        shape = size
    if requires_grad:
        return paddle.randn(shape, dtype).requires_grad_(True)
    else:
        return paddle.randn(shape, dtype)


def randn_like(input,
               dtype=None,
               layout=None,
               device=None,
               requires_grad=False,
               memory_format=None):
    shape = input.shape
    if requires_grad:
        return paddle.randn(shape, dtype).requires_grad_(True)
    else:
        return paddle.randn(shape, dtype)


def randperm(n,
             *,
             generator=None,
             out=None,
             dtype="int64",
             layout=None,
             device=None,
             requires_grad=False,
             pin_memory=False):
    if requires_grad:
        return paddle.randperm(n, dtype).requires_grad_(True)
    else:
        return paddle.randperm(n, dtype)


def save(obj, f, pickle_module=None, pickle_protocol=2):
    return paddle.save(obj.state_dict(), f, pickle_protocol=pickle_protocol)


def split(tensor, split_size_or_sections, dim=0):
    return paddle.split(tensor, split_size_or_sections, dim)


def sqrt(input, *, out=None):
    return paddle.sqrt(input)


def stack(tensors, dim=0, *, out=None):
    return paddle.stack(tensors, dim)


def sum(input, dim=None, keepdim=False, *, out=None):
    if dim is None:
        warnings.warn('The output of paddle.sum is not scalar!')
        return paddle.sum(input)
    else:
        return paddle.sum(input, axis=dim, keepdim=keepdim)


def unsqueeze(input, dim):
    return paddle.squeeze(input, dim)


def zeros(*size,
          out=None,
          dtype=None,
          layout=None,
          device=None,
          requires_grad=False):
    if len(size) == 1 and isinstance(size[0], (tuple, list)):
        shape = size[0]
    else:
        shape = size
    if requires_grad:
        return paddle.zeros(shape, dtype).requires_grad_(True)
    else:
        return paddle.zeros(shape, dtype)


def zeros_like(input,
               *,
               dtype=None,
               layout=None,
               device=None,
               requires_grad=False,
               memory_format=None):
    if requires_grad:
        return paddle.zeros_like(input, dtype).requires_grad_(True)
    else:
        return paddle.zeros_like(input, dtype)


class DataParallel(paddle.DataParallel):
    def __init__(self, module, device_ids=None, output_device=None, dim=0):
        super().__init__(module)


def invalid(*args, **kwargs):
    return None



# -*- coding: UTF-8 -*-
# @Time    : 2022/9/11 23:49
# @File    : config.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from typing import Union, List
from antgo.pipeline.hparam import param_scope


class ConfigMixin:
    """
    Mixin to config DC, such as set the `parallel`, `chunksize`, `jit`.
    """

    def __init__(self) -> None:
        super().__init__()
        with param_scope() as hp:
            parent = hp().data_collection.parent(None)
        if parent is not None and hasattr(parent, '_config'):
            self._config = parent._config
        else:
            self._config = None
        if parent is None or not hasattr(parent, '_num_worker'):
            self._num_worker = None
        if parent is None or not hasattr(parent, '_chunksize'):
            self._chunksize = None
        if parent is None or not hasattr(parent, '_jit'):
            self._jit = None
        if parent is None or not hasattr(parent, '_format_priority'):
            self._format_priority = None

    def config(self, parallel: int = None, chunksize: int = None, jit: Union[str, dict] = None, format_priority: List[str] = None):
        """
        Set the parameters in DC.

        Args:
            parallel (`int`):
               Set the number of parallel execution for following calls.
            chunksize (`int`):
               Set the chunk size for arrow.
            jit (`Union[str, dict]`):
               It can set to "numba", this mode will speed up the Operator's function, but it may also need to return to python mode due to JIT
               failure, which will take longer, so please set it carefully.
            format_priority (`List[str]`):
                The priority list of format.
        """
        dc = self
        if jit is not None:
            dc = dc.set_jit(compiler=jit)
        if parallel is not None:
            dc = dc.set_parallel(num_worker=parallel)
        if chunksize is not None:
            dc = dc.set_chunksize(chunksize=chunksize)
        if format_priority is not None:
            dc = dc.set_format_priority(format_priority=format_priority)
        return dc

    def get_config(self):
        """
        Return the config in DC, such as `parallel`, `chunksize`, `jit` and `format_priority`.
        """
        self._config = {}

        if hasattr(self, '_num_worker'):
            self._config['parallel'] = self._num_worker
        if hasattr(self, '_chunksize'):
            self._config['chunksize'] = self._chunksize
        if hasattr(self, '_jit'):
            self._config['jit'] = self._jit
        if hasattr(self, '_format_priority'):
            self._config['format_priority'] = self._format_priority
        return self._config

    def pipeline_config(self, parallel: int = None, chunksize: int = None, jit: Union[str, dict] = None, format_priority: List[str] = None):
        """
        Set the parameters in DC.

        Args:
            parallel (`int`):
               Set the number of parallel execution for following calls.
            chunksize (`int`):
               Set the chunk size for arrow.
            jit (`Union[str, dict]`):
               It can set to "numba", this mode will speed up the Operator's function, but it may also need to return to python mode due to JIT
               failure, which will take longer, so please set it carefully.
            format_priority (`List[str]`):
                The priority list of format.
        """
        dc = self
        if jit is not None:
            dc = dc.set_jit(compiler=jit)
        if parallel is not None:
            dc = dc.set_parallel(num_worker=parallel)
        if chunksize is not None:
            dc = dc.set_chunksize(chunksize=chunksize)
        if format_priority is not None:
            dc = dc.set_format_priority(format_priority=format_priority)
        return dc

    def get_pipeline_config(self):
        """
        Return the config in DC, such as `parallel`, `chunksize`, `jit` and `format_priority`.
        """
        self._pipeline_config = {}

        if hasattr(self, '_num_worker'):
            self._pipeline_config['parallel'] = self._num_worker
        if hasattr(self, '_chunksize'):
            self._pipeline_config['chunksize'] = self._chunksize
        if hasattr(self, '_jit'):
            self._pipeline_config['jit'] = self._jit
        if hasattr(self, '_format_priority'):
            self._pipeline_config['format_priority'] = self._format_priority
        return self._pipeline_config



import warnings

import torch

from antgo.framework.helper.utils import digit_version


def is_jit_tracing() -> bool:
    if (torch.__version__ != 'parrots'
            and digit_version(torch.__version__) >= digit_version('1.6.0')):
        on_trace = torch.jit.is_tracing()
        # In PyTorch 1.6, torch.jit.is_tracing has a bug.
        # Refers to https://github.com/pytorch/pytorch/issues/42448
        if isinstance(on_trace, bool):
            return on_trace
        else:
            return torch._C._is_tracing()
    else:
        warnings.warn(
            'torch.jit.is_tracing is only supported after v1.6.0. '
            'Therefore is_tracing returns False automatically. Please '
            'set on_trace manually if you are using trace.', UserWarning)
        return False



# Copyright (c) OpenMMLab. All rights reserved.
import warnings

from antgo.framework.helper.utils import Registry, build_from_cfg

PRIOR_GENERATORS = Registry('Generator for anchors and points')

ANCHOR_GENERATORS = PRIOR_GENERATORS


def build_prior_generator(cfg, default_args=None):
    return build_from_cfg(cfg, PRIOR_GENERATORS, default_args)


def build_anchor_generator(cfg, default_args=None):
    warnings.warn(
        '``build_anchor_generator`` would be deprecated soon, please use '
        '``build_prior_generator`` ')
    return build_prior_generator(cfg, default_args=default_args)



# -*- coding: UTF-8 -*-
# @Time    : 2022/9/21 22:42
# @File    : inference_onnx_op.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

from antgo.pipeline.engine import *
from antgo.pipeline.models.utils.utils import *
from antgo.interactcontext import InteractContext
import torch
import numpy as np
import onnx
import onnxruntime as ort
from multiprocessing import Process
from multiprocessing import queues
import logging
import os


@register
class inference_onnx_op(object):
    def __init__(self, onnx_path, input_fields, device_id=-1, **kwargs):
        self.sess = ort.InferenceSession(onnx_path)
        if device_id >= 0:
            self.sess.set_providers(['CUDAExecutionProvider'], [{'device_id': device_id}])
        self.input_fields = input_fields
        self.mean_val = kwargs.get('mean', None)   # 均值
        self.std_val = kwargs.get('std', None)     # 方差
        self.rgb2bgr = kwargs.get('rgb2bgr', False)

    def __call__(self, *args):        
        input_map = {}
        for field, data in zip(self.input_fields, args):
            if self.mean_val is not None:
                if len(data.shape) == 4:
                    # NxHxWx3
                    if self.rgb2bgr:
                        data = data[:,:,:,::-1]
                    data = data - np.reshape(np.array(self.mean_val), (1,1,1,3))
                    data = data / np.reshape(np.array(self.std_val), (1,1,1,3))

                    # Nx3xHxW
                    data = np.transpose(data, (0,3,1,2))
                else:
                    # HxWx3
                    if self.rgb2bgr:
                        data = data[:,:,::-1]
                    data = data - np.reshape(np.array(self.mean_val), (1,1,3))
                    data = data / np.reshape(np.array(self.std_val), (1,1,3))

                    # -> 1xHxWx3
                    data = np.expand_dims(data, 0)
                    # -> 1x3xHxW
                    data = np.transpose(data, (0,3,1,2))

            if data.dtype != np.float32:
                data = data.astype(np.float32)
            input_map[field] = data

        output = self.sess.run(None, input_map)
        if isinstance(output, list) or isinstance(output, tuple):
            return tuple(output)
        
        return (output,)
    

def __session_run_in_process(onnx_path, device_id, input_fields, input_queue, output_queue):
    sess = ort.InferenceSession(onnx_path)
    sess.set_providers(['CUDAExecutionProvider'], [{'device_id': device_id}])

    while True:
        data = input_queue.get()
        if data is None:
            break
        
        input_map = {}
        for field, data in zip(input_fields, data):
            data = np.transpose(data, [2,0,1])
            input_map[field] = data[np.newaxis,:,:,:].astype(np.float32)
            
        output = sess.run(None, input_map)
        if isinstance(output, list) or isinstance(output, tuple):
            output_queue.put(tuple(output))
            continue
        
        output_queue.put(tuple(output))


@register
class ensemble_onnx_op(object):
    def __init__(self, onnx_path_list, input_fields, device_ids) -> None:
        self.input_fields = input_fields
        assert(self.backend in ['process', 'thread'])
        
        self.input_queues =[queues.Queue() for _ in range(len(onnx_path_list))]
        self.output_queues = [queues.Queue() for _ in range(len(onnx_path_list))]
        self.processes = []
        self.multi_process(self.onnx_path_list, device_ids, self.input_queues, self.output_queues)

    def multi_process(self, onnx_path_list, device_ids, input_queues, output_queues):
        self.processes = []
        for onnx_path, device_id in zip(onnx_path_list, device_ids):
            process = Process(target=__session_run_in_process, args=(onnx_path, device_id, self.input_fields, input_queues, output_queues))
            process.daemon = True
            process.start()
            self.processes.append(process)

    def __call__(self, *args):
        # pass
        for input_queue in self.input_queues:
            input_queue.put(args)
        
        ensemble_result = []
        for output_queue in self.output_queues:
            result = output_queue.get()
            if len(ensemble_result) == 0:
                ensemble_result = [[] for _ in range(len(result))]
            for i, v in enumerate(result):
                ensemble_result[i].append(v)

        out = []
        for i, v in enumerate(ensemble_result):
            out.append(np.mean(np.stack(v, 0), axis=0, keepdims=False))
        return tuple(out)


@register
class ensemble_shell_op(object):
    def __init__(self, shell_script_list, parse_fields=[], device_ids=None):
        self.parse_fields = parse_fields
        self.device_ids = device_ids
        self.ctx = InteractContext()
        self.ctx.ensemble.start("A", config = {
                'weight': 1.0,              # ignore
                'role': 'master',
                'worker': len(shell_script_list),
                'stage': 'merge',
                'feedback': True,
                'background': False})        

        for shell_script in shell_script_list:
            logging.info(f'Shell script {shell_script}')
            process = Process(target=os.system, args=(shell_script,))
            process.daemon = True
            process.start()
    
    def __call__(self, *args):
        # 推送到聚合服务
        if len(args) == 1:
            self.ctx.recorder.put(args[0])
        else:
            self.ctx.recorder.put(args)
        
        # 获得聚合结果
        result = self.ctx.recorder.avg()
        # k,v
        ss = []
        for field in self.parse_fields:
            ss.append(result[field])
        return ss


# Copyright (c) OpenMMLab. All rights reserved.
import warnings

import torch
from torch.nn import GroupNorm, LayerNorm

from antgo.framework.helper.utils import _BatchNorm, _InstanceNorm, build_from_cfg, is_list_of
from antgo.framework.helper.utils.ext_loader import check_ops_exist
from .builder import OPTIMIZER_BUILDERS, OPTIMIZERS


@OPTIMIZER_BUILDERS.register_module()
class DefaultOptimizerConstructor:
    """Default constructor for optimizers.

    By default each parameter share the same optimizer settings, and we
    provide an argument ``paramwise_cfg`` to specify parameter-wise settings.
    It is a dict and may contain the following fields:

    - ``custom_keys`` (dict): Specified parameters-wise settings by keys. If
      one of the keys in ``custom_keys`` is a substring of the name of one
      parameter, then the setting of the parameter will be specified by
      ``custom_keys[key]`` and other setting like ``bias_lr_mult`` etc. will
      be ignored. It should be noted that the aforementioned ``key`` is the
      longest key that is a substring of the name of the parameter. If there
      are multiple matched keys with the same length, then the key with lower
      alphabet order will be chosen.
      ``custom_keys[key]`` should be a dict and may contain fields ``lr_mult``
      and ``decay_mult``. See Example 2 below.
    - ``bias_lr_mult`` (float): It will be multiplied to the learning
      rate for all bias parameters (except for those in normalization
      layers and offset layers of DCN).
    - ``bias_decay_mult`` (float): It will be multiplied to the weight
      decay for all bias parameters (except for those in
      normalization layers, depthwise conv layers, offset layers of DCN).
    - ``norm_decay_mult`` (float): It will be multiplied to the weight
      decay for all weight and bias parameters of normalization
      layers.
    - ``dwconv_decay_mult`` (float): It will be multiplied to the weight
      decay for all weight and bias parameters of depthwise conv
      layers.
    - ``dcn_offset_lr_mult`` (float): It will be multiplied to the learning
      rate for parameters of offset layer in the deformable convs
      of a model.
    - ``bypass_duplicate`` (bool): If true, the duplicate parameters
      would not be added into optimizer. Default: False.

    Note:

        1. If the option ``dcn_offset_lr_mult`` is used, the constructor will
        override the effect of ``bias_lr_mult`` in the bias of offset layer.
        So be careful when using both ``bias_lr_mult`` and
        ``dcn_offset_lr_mult``. If you wish to apply both of them to the offset
        layer in deformable convs, set ``dcn_offset_lr_mult`` to the original
        ``dcn_offset_lr_mult`` * ``bias_lr_mult``.

        2. If the option ``dcn_offset_lr_mult`` is used, the constructor will
        apply it to all the DCN layers in the model. So be careful when the
        model contains multiple DCN layers in places other than backbone.

    Args:
        model (:obj:`nn.Module`): The model with parameters to be optimized.
        optimizer_cfg (dict): The config dict of the optimizer.
            Positional fields are

                - `type`: class name of the optimizer.

            Optional fields are

                - any arguments of the corresponding optimizer type, e.g.,
                  lr, weight_decay, momentum, etc.
        paramwise_cfg (dict, optional): Parameter-wise options.

    Example 1:
        >>> model = torch.nn.modules.Conv1d(1, 1, 1)
        >>> optimizer_cfg = dict(type='SGD', lr=0.01, momentum=0.9,
        >>>                      weight_decay=0.0001)
        >>> paramwise_cfg = dict(norm_decay_mult=0.)
        >>> optim_builder = DefaultOptimizerConstructor(
        >>>     optimizer_cfg, paramwise_cfg)
        >>> optimizer = optim_builder(model)

    Example 2:
        >>> # assume model have attribute model.backbone and model.cls_head
        >>> optimizer_cfg = dict(type='SGD', lr=0.01, weight_decay=0.95)
        >>> paramwise_cfg = dict(custom_keys={
                'backbone': dict(lr_mult=0.1, decay_mult=0.9)})
        >>> optim_builder = DefaultOptimizerConstructor(
        >>>     optimizer_cfg, paramwise_cfg)
        >>> optimizer = optim_builder(model)
        >>> # Then the `lr` and `weight_decay` for model.backbone is
        >>> # (0.01 * 0.1, 0.95 * 0.9). `lr` and `weight_decay` for
        >>> # model.cls_head is (0.01, 0.95).
    """

    def __init__(self, optimizer_cfg, paramwise_cfg=None):
        if not isinstance(optimizer_cfg, dict):
            raise TypeError('optimizer_cfg should be a dict',
                            f'but got {type(optimizer_cfg)}')
        self.optimizer_cfg = optimizer_cfg
        self.paramwise_cfg = {} if paramwise_cfg is None else paramwise_cfg
        self.base_lr = optimizer_cfg.get('lr', None)
        self.base_wd = optimizer_cfg.get('weight_decay', None)
        self._validate_cfg()

    def _validate_cfg(self):
        if not isinstance(self.paramwise_cfg, dict):
            raise TypeError('paramwise_cfg should be None or a dict, '
                            f'but got {type(self.paramwise_cfg)}')

        if 'custom_keys' in self.paramwise_cfg:
            if not isinstance(self.paramwise_cfg['custom_keys'], dict):
                raise TypeError(
                    'If specified, custom_keys must be a dict, '
                    f'but got {type(self.paramwise_cfg["custom_keys"])}')
            if self.base_wd is None:
                for key in self.paramwise_cfg['custom_keys']:
                    if 'decay_mult' in self.paramwise_cfg['custom_keys'][key]:
                        raise ValueError('base_wd should not be None')

        # get base lr and weight decay
        # weight_decay must be explicitly specified if mult is specified
        if ('bias_decay_mult' in self.paramwise_cfg
                or 'norm_decay_mult' in self.paramwise_cfg
                or 'dwconv_decay_mult' in self.paramwise_cfg):
            if self.base_wd is None:
                raise ValueError('base_wd should not be None')

    def _is_in(self, param_group, param_group_list):
        assert is_list_of(param_group_list, dict)
        param = set(param_group['params'])
        param_set = set()
        for group in param_group_list:
            param_set.update(set(group['params']))

        return not param.isdisjoint(param_set)

    def add_params(self, params, module, prefix='', is_dcn_module=None):
        """Add all parameters of module to the params list.

        The parameters of the given module will be added to the list of param
        groups, with specific rules defined by paramwise_cfg.

        Args:
            params (list[dict]): A list of param groups, it will be modified
                in place.
            module (nn.Module): The module to be added.
            prefix (str): The prefix of the module
            is_dcn_module (int|float|None): If the current module is a
                submodule of DCN, `is_dcn_module` will be passed to
                control conv_offset layer's learning rate. Defaults to None.
        """
        # get param-wise options
        custom_keys = self.paramwise_cfg.get('custom_keys', {})
        # first sort with alphabet order and then sort with reversed len of str
        sorted_keys = sorted(sorted(custom_keys.keys()), key=len, reverse=True)

        bias_lr_mult = self.paramwise_cfg.get('bias_lr_mult', 1.)
        bias_decay_mult = self.paramwise_cfg.get('bias_decay_mult', 1.)
        norm_decay_mult = self.paramwise_cfg.get('norm_decay_mult', 1.)
        dwconv_decay_mult = self.paramwise_cfg.get('dwconv_decay_mult', 1.)
        bypass_duplicate = self.paramwise_cfg.get('bypass_duplicate', False)
        dcn_offset_lr_mult = self.paramwise_cfg.get('dcn_offset_lr_mult', 1.)

        # special rules for norm layers and depth-wise conv layers
        is_norm = isinstance(module,
                             (_BatchNorm, _InstanceNorm, GroupNorm, LayerNorm))
        is_dwconv = (
            isinstance(module, torch.nn.Conv2d)
            and module.in_channels == module.groups)

        for name, param in module.named_parameters(recurse=False):
            param_group = {'params': [param]}
            if not param.requires_grad:
                params.append(param_group)
                continue
            if bypass_duplicate and self._is_in(param_group, params):
                warnings.warn(f'{prefix} is duplicate. It is skipped since '
                              f'bypass_duplicate={bypass_duplicate}')
                continue
            # if the parameter match one of the custom keys, ignore other rules
            is_custom = False
            for key in sorted_keys:
                if key in f'{prefix}.{name}':
                    is_custom = True
                    lr_mult = custom_keys[key].get('lr_mult', 1.)
                    param_group['lr'] = self.base_lr * lr_mult
                    if self.base_wd is not None:
                        decay_mult = custom_keys[key].get('decay_mult', 1.)
                        param_group['weight_decay'] = self.base_wd * decay_mult
                    break

            if not is_custom:
                # bias_lr_mult affects all bias parameters
                # except for norm.bias dcn.conv_offset.bias
                if name == 'bias' and not (is_norm or is_dcn_module):
                    param_group['lr'] = self.base_lr * bias_lr_mult

                if (prefix.find('conv_offset') != -1 and is_dcn_module
                        and isinstance(module, torch.nn.Conv2d)):
                    # deal with both dcn_offset's bias & weight
                    param_group['lr'] = self.base_lr * dcn_offset_lr_mult

                # apply weight decay policies
                if self.base_wd is not None:
                    # norm decay
                    if is_norm:
                        param_group[
                            'weight_decay'] = self.base_wd * norm_decay_mult
                    # depth-wise conv
                    elif is_dwconv:
                        param_group[
                            'weight_decay'] = self.base_wd * dwconv_decay_mult
                    # bias lr and decay
                    elif name == 'bias' and not is_dcn_module:
                        # TODO: current bias_decay_mult will have affect on DCN
                        param_group[
                            'weight_decay'] = self.base_wd * bias_decay_mult
            params.append(param_group)

        if check_ops_exist():
            from mmcv.ops import DeformConv2d, ModulatedDeformConv2d
            is_dcn_module = isinstance(module,
                                       (DeformConv2d, ModulatedDeformConv2d))
        else:
            is_dcn_module = False
        for child_name, child_mod in module.named_children():
            child_prefix = f'{prefix}.{child_name}' if prefix else child_name
            self.add_params(
                params,
                child_mod,
                prefix=child_prefix,
                is_dcn_module=is_dcn_module)

    def __call__(self, model):
        if hasattr(model, 'module'):
            model = model.module

        optimizer_cfg = self.optimizer_cfg.copy()
        # if no paramwise option is specified, just use the global setting
        if not self.paramwise_cfg:
            optimizer_cfg['params'] = model.parameters()
            return build_from_cfg(optimizer_cfg, OPTIMIZERS)

        # set param-wise lr and weight decay recursively
        params = []
        self.add_params(params, model)
        optimizer_cfg['params'] = params

        return build_from_cfg(optimizer_cfg, OPTIMIZERS)



# -*- coding: UTF-8 -*-
# @Time    : 2022/5/2 12:33
# @File    : ops.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import torch
from antgo.framework.paddle2torch.utils import *


class ParamAttr(object):
  def __init__(self,
               name=None,
               initializer=None,
               learning_rate=1.0,
               regularizer=None,
               trainable=True,
               do_model_average=False,
               need_clip=True):
    self.name = name
    self.initializer = initializer
    self.learning_rate = learning_rate
    self.regularizer = regularizer
    self.trainable = trainable
    self.do_model_average = do_model_average
    self.need_clip = need_clip


def reshape(x, shape):
  return torch.reshape(x, shape)


def add(input, other):
  return torch.add(input, other)

def clip(input, min, max):
  return torch.clip(input, min, max)

def concat(tensors, axis=0):
  return torch.concat(tensors, dim=axis)

def exp(input):
  return torch.exp(input)

def log(input):
  return torch.log(input)

def logical_and(input, other):
  return torch.logical_and(input, other)

def logical_not(input):
  return torch.logical_not(input)

def logical_or(input, other):
  return torch.logical_or(input, other)

def logical_xor(input, other):
  return torch.logical_xor(input, other)


def matmul(input, other):
  return torch.matmul(input, other)


def multiply(input, other):
  return torch.mul(input, other)

def maximum(input, other):
  return torch.max(input, other)

def max(x, axis=None, keepdim=False, name=None):
  return torch.max(x)

def mean(x, axis=-1, keepdim=False, name=None):
  return torch.mean(x)

def min(x, axis=-1, keepdim=False, name=None):
  return torch.min(x)

def ones(shape, dtype):
  return torch.ones(shape, dtype)

def ones_like(input):
  return torch.ones_like(input)

def split(x, num_or_sections, axis=0, name=None):
  return torch.split(x,num_or_sections,axis)

def sqrt(x, name=None):
  return torch.sqrt(x)


def stack(x, axis=0, name=None):
  return torch.stack(x,dim=axis)

def sum(x, axis=None, dtype=None, keepdim=False, name=None):
  return torch.sum(x,dim=axis,keepdim=keepdim)

def squeeze(x, axis=None, name=None):
  return torch.unsqueeze(x,dim=axis)

def zeros(shape, dtype=None, name=None):
  return torch.zeros(shape, dtype=dtype)

def zeros_like(x, dtype=None, name=None):
  return torch.zeros_like(x)

def uniform(shape, dtype='float32', min=-1.0, max=1.0, seed=0, name=None):
  x = torch.rand(shape, dtype=PADDLE_TORCH_TYPE_MAPPER[dtype])
  x = x * (max - min) + min
  return x

def rand(shape, dtype='float32'):
  return torch.rand(shape, dtype=PADDLE_TORCH_TYPE_MAPPER[dtype])


def to_tensor(data, dtype=None, place=None, stop_gradient=True):
  return torch.tensor(data,
                      PADDLE_TORCH_TYPE_MAPPER[dtype],
                      device=place if place is None else place.device(),
                      requires_grad=not stop_gradient)

def flatten(x, start_axis=0, stop_axis=- 1, name=None):
  return torch.flatten(x,start_axis,stop_axis)

def save(obj, f):
  return torch.save(obj, f)


def load(path, **configs):
  return torch.load(path,map_location=configs.get('map_location', None))




# -*- coding: UTF-8 -*-
# @Time    : 2022/9/12 17:48
# @File    : operator_loader.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

import importlib
import sys
import os
import subprocess
from pathlib import Path
from typing import Any, List, Dict, Union

from antgo.pipeline.operators import Operator
from antgo.pipeline.operators.nop import NOPOperator
from antgo.pipeline.engine import *

from antgo.pipeline.hparam import param_scope


class OperatorLoader:
    """Wrapper class used to load operators from either local cache or a remote
    location.

    Args:
        cache_path: (`str`)
            Local cache path to use. If not specified, it will default to
            `$HOME/.antgo/operators`.
    """
    def __init__(self, cache_path: str = None):
        if cache_path is None:
            self._cache_path = LOCAL_OPERATOR_CACHE
        else:
            self._cache_path = Path(cache_path)

    def _load_interal_op(self, op_name: str, arg: List[Any], kws: Dict[str, Any]):
        if op_name in ['_start_op', '_end_op']:
            return NOPOperator()
        elif op_name == '_concat':
            return ConcatOperator(*arg, **kws)
        else:
            # Not a interal operators
            return None

    def load_operator_from_internal(self, function: str, arg: List[Any], kws: Dict[str, Any], tag: str) -> Operator:  # pylint: disable=unused-argument
        return self._load_interal_op(function, arg, kws)

    def load_operator_from_registry(self, function: str, arg: List[Any], kws: Dict[str, Any], tag: str) -> Operator:  # pylint: disable=unused-argument
        op = OperatorRegistry.resolve(function)
        return self.instance_operator(op, arg, kws) if op is not None else None

    def load_operator_from_mm(self, function: str, arg: List[Any], kws: Dict[str, Any], tag: str) -> Operator:  # pylint: disable=unused-argument
        module, fname = function.split('/')
        op_cls = ''.join(x.capitalize() or '_' for x in fname.split('_'))
        module = '.'.join(['antgo.pipeline.hub.external', '{}.{}'.format(module, fname)])

        try:
            op = getattr(importlib.import_module(module), op_cls)
        except:
            subprocess.check_call([sys.executable, '-m', 'pip3', 'install', 'openmim'])
            subprocess.check_call([sys.executable, '-m', 'pip3', 'install', 'mmengine'])
            subprocess.check_call([sys.executable, '-m', 'pip3', 'install', 'mmcv-full'])
            if fname == 'detector':
                subprocess.check_call([sys.executable, '-m', 'pip3', 'install', 'mmdet'])
            elif fname == 'segmentor':
                subprocess.check_call([sys.executable, '-m', 'pip3', 'install', 'mmsegmentation'])
            elif fname == 'classification':
                subprocess.check_call([sys.executable, '-m', 'pip3', 'install', 'mmcls'])
            elif fname == 'ocr':
                subprocess.check_call([sys.executable, '-m', 'pip3', 'install', 'mmocr'])
            elif fname == 'pose':
                subprocess.check_call([sys.executable, '-m', 'pip3', 'install', 'mmpose'])
            elif fname == 'editing':
                subprocess.check_call([sys.executable, '-m', 'git', 'clone', 'https://github.com/open-mmlab/mmediting.git'])
                subprocess.check_call([sys.executable, '-m', 'cd', 'mmediting'])
                subprocess.check_call([sys.executable, '-m', 'pip3', 'install', '-e', '.'])

            op = getattr(importlib.import_module(module), op_cls)

        return self.instance_operator(op, arg, kws) if op is not None else None

    def load_operator_from_deploy(self, function: str, arg: List[Any], kws: Dict[str, Any], tag: str) -> Operator:  # pylint: disable=unused-argument
        module, fname = function.split('/')
        if module != 'deploy':
            return None

        op = getattr(importlib.import_module('antgo.pipeline.deploy.cpp_op'), 'CppOp', None)
        if op is None:
            return None
        kws.update({
            'func_op_name': fname.replace('-', '_')
        })
        return self.instance_operator(op, arg, kws) if op is not None else None

    def load_operator_from_eagleeye(self, function: str, arg: List[Any], kws: Dict[str, Any], tag: str) -> Operator:  # pylint: disable=unused-argument
        module, fname = function.split('/')
        if module != 'eagleeye':
            return None

        os.makedirs('.3rd', exist_ok=True)
        if not os.path.exists(os.path.join('.3rd', 'eagleeye-py')):
            if not os.path.exists(os.path.join('.3rd', 'eagleeye')):
                os.system('cd .3rd/ && git clone https://github.com/jianzfb/eagleeye.git')
                        
            if 'darwin' in sys.platform:
                os.system('cd .3rd/eagleeye && bash osx_build.sh BUILD_PYTHON_MODULE && cp -r install ../eagleeye-py')
            else:
                first_comiple = False
                if not os.path.exists(os.path.join('.3rd', 'eagleeye-py')):
                    first_comiple = True
                os.system('cd .3rd/eagleeye && bash linux_build.sh BUILD_PYTHON_MODULE && cp -r install ../eagleeye-py')
                if first_comiple:
                    # 增加搜索.so路径
                    cur_abs_path = os.path.abspath(os.curdir)
                    so_abs_path = os.path.join(cur_abs_path, ".3rd/eagleeye-py/libs/X86-64")
                    os.system(f'echo "{so_abs_path}" >> /etc/ld.so.conf && ldconfig')

        op = getattr(importlib.import_module('antgo.pipeline.eagleeye.core_op'), 'CoreOp', None)
        if op is None:
            return None
        kws.update({
            'func_op_name': fname.replace('-', '_')
        })
        return self.instance_operator(op, arg, kws) if op is not None else None


    def load_operator_from_packages(self, function: str, arg: List[Any], kws: Dict[str, Any], tag: str) -> Operator:  # pylint: disable=unused-argument
        try:
            module, fname = function.split('/')
            load_func = getattr(self, f'load_operator_from_{module}', None)
            if load_func is not None:
                return load_func(function, arg, kws, tag)

            fname = fname.replace('-', '_')
            op_cls = ''.join(x.capitalize() or '_' for x in fname.split('_'))

            # module = '.'.join([module, fname, fname])
            module = '.'.join(['antgo.pipeline.models', '{}.{}'.format(module, fname), fname])
            op = getattr(importlib.import_module(module), op_cls)
            return self.instance_operator(op, arg, kws) if op is not None else None
        except Exception:  # pylint: disable=broad-except
            return None

    def load_operator_from_remote(self, function: str, arg: List[Any], kws: Dict[str, Any], tag: str) -> Operator:
        # 使用云端API
        return None

    def load_operator(self, function: str, arg: List[Any], kws: Dict[str, Any], tag: str) -> Operator:
        """Attempts to load an operators from cache. If it does not exist, looks up the
        operators in a remote location and downloads it to cache instead. By standard
        convention, the operators must be called `Operator` and all associated data must
        be contained within a single directory.

        Args:
            function: (`str`)
                Origin and method/class name of the operators. Used to look up the proper
                operators in cache.
        Raises:
            FileExistsError
                Cannot find operators.
        """

        for factory in [self.load_operator_from_internal,
                        self.load_operator_from_registry,
                        self.load_operator_from_remote,
                        self.load_operator_from_packages,
                        self.load_operator_from_deploy,
                        self.load_operator_from_eagleeye]:
            op = factory(function, arg, kws, tag)
            if op is not None:
                return op
        return None

    def instance_operator(self, op, arg: List[Any], kws: Dict[str, Any]) -> Operator:
        with param_scope() as hp:
            return op(*arg, **kws) if kws is not None else op()



from collections import defaultdict
import re
import os
import queue

_INCLUDE_FILE_REG = re.compile(r'^\s*#include\s*(?:"|<)\s*(.*?)\s*(?:"|>)\s*')

_C_EXTENSIONS = ['.cpp', '.c', '.cu']


def _is_c_file(fname):
    return os.path.splitext(fname)[-1] in _C_EXTENSIONS


def get_include_file(fname):
    '''return a list of all inclued filename'''
    res = []
    for line in open(fname):
        u = _INCLUDE_FILE_REG.search(line)
        if u is not None:
            inc_fname = u.groups()[0]
            res.append(inc_fname)
    return res


def get_dependant_graph(fname):
    '''return a partial dependant graph, a dict of (str, a list of str)
    included file -> code file
    '''
    graph = defaultdict(list)
    vis = set()
    if not _is_c_file(fname):
        return graph
    fname = os.path.abspath(fname)
    q = queue.Queue()
    q.put(fname)
    dirname = os.path.dirname(fname)
    while not q.empty():
        name = q.get()
        if name in vis:
            continue
        vis.add(name)
        for iname in get_include_file(fname):
            iname = os.path.abspath(os.path.join(dirname, iname))
            graph[iname].append(fname)
            q.put(iname)
    return graph



# -*- coding: UTF-8 -*-
# @Time    : 18-3-19
# @File    : ali_fashion_landmark_ne.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.task.task import *
from antgo.measures.base import *
import numpy as np

default = {'AntALIFashionLandmarkNE': ('FashionLandmarkNE', 'LANDMARK')}
class AntALIFashionLandmarkNE(AntMeasure):
  def __init__(self, task):
    super(AntALIFashionLandmarkNE, self).__init__(task, 'FashionLandmarkNE')
    assert(task.task_type == 'LANDMARK')
    self.is_support_rank = True
    self.is_inverse = True
    
  def eva(self, data, label):
    '''
    
    :param data:
    :param label:
    :return:
    '''
    
    if label is not None:
      data = zip(data, label)
    
    measure_val = 0.0
    sample_scores = []
    for predict, gt in data:
      predict_landmarks, predict_category = predict

      gt_landmarks = gt['landmark']
      gt_category = gt['category']
      id = gt['id']

      assert(predict_category == gt_category)
      
      # normalized parameter
      sk = 0.0
      if gt_category in ['blouse', 'outwear', 'dress']:
        # 衬衫，外套，连衣裙
        if gt_landmarks[5][4] in [0, 1] and gt_landmarks[6][4] in [0, 1]:
          armpit_left_px = gt_landmarks[5][2]
          armpit_left_py = gt_landmarks[5][3]

          armpit_right_px = gt_landmarks[6][2]
          armpit_right_py = gt_landmarks[6][3]

          sk = np.sqrt(np.power(armpit_left_px - armpit_right_px,2)+
                       np.power(armpit_left_py - armpit_right_py,2)) + 1e-6
      else:
        # 裤子，半身裙
        if gt_landmarks[15][4] in [0, 1] and gt_landmarks[16][4] in [0, 1]:
          waistband_left_px = gt_landmarks[15][2]
          waistband_left_py = gt_landmarks[15][3]

          waistband_right_px = gt_landmarks[16][2]
          waistband_right_py = gt_landmarks[16][3]

          sk = np.sqrt(np.power(waistband_left_px - waistband_right_px, 2) +
                       np.power(waistband_left_py - waistband_right_py, 2)) + 1e-6

      if sk < 1e-5:
        continue

      denominator = 1e-6
      numerator = 0.0
      for predict_landmark, gt_landmark in zip(predict_landmarks, gt_landmarks):
        predict_x, predict_y, predict_visible = predict_landmark
        gt_x, gt_y, gt_visible = gt_landmark[2:]
        
        if gt_visible == 1:
          denominator += 1
          numerator += (np.sqrt(np.power(predict_x-gt_x,2.0) + np.power(predict_y-gt_y,2.0))) / sk if predict_visible == 1 else 1.0

      score = numerator / denominator
      sample_scores.append({'id': id, 'score': score, 'category': gt_category})
      measure_val += score
    
    measure_val /= float(len(sample_scores))
    return {'statistic': {'name': self.name,
                          'value':[{'name': self.name, 'value': measure_val, 'type': 'SCALAR'}]},
            'info': sample_scores}


import torch.nn.functional as F
from torch import nn
import torch

from antgo.framework.helper.models.builder import MODELS, build_model
from antgo.framework.helper.multi_stream_module import MultiSteamModule
from antgo.framework.helper.models.builder import build_loss


class ABF(nn.Module):
    def __init__(self, in_channel, mid_channel, out_channel, fuse):
        super(ABF, self).__init__()
        self.conv1 = nn.Sequential(
            nn.Conv2d(in_channel, mid_channel, kernel_size=1, bias=False),
            nn.BatchNorm2d(mid_channel),
        )
        self.conv2 = nn.Sequential(
            nn.Conv2d(mid_channel, out_channel,kernel_size=3,stride=1,padding=1,bias=False),
            nn.BatchNorm2d(out_channel),
        )
        if fuse:
            self.att_conv = nn.Sequential(
                    nn.Conv2d(mid_channel*2, 2, kernel_size=1),
                    nn.Sigmoid(),
                )
        else:
            self.att_conv = None
        nn.init.kaiming_uniform_(self.conv1[0].weight, a=1)  # pyre-ignore
        nn.init.kaiming_uniform_(self.conv2[0].weight, a=1)  # pyre-ignore

    def forward(self, x, y=None, shape=None, out_shape=None):
        n,_,h,w = x.shape
        # transform student features
        x = self.conv1(x)
        if self.att_conv is not None:
            # upsample residual features
            y = F.interpolate(y, (shape,shape), mode="nearest")
            # fusion
            z = torch.cat([x, y], dim=1)
            z = self.att_conv(z)
            x = (x * z[:,0].view(n,1,h,w) + y * z[:,1].view(n,1,h,w))
        # output 
        if x.shape[-1] != out_shape:
            x = F.interpolate(x, (out_shape, out_shape), mode="nearest")
        y = self.conv2(x)
        return y, x


@MODELS.register_module()
class ReviewKD(MultiSteamModule):
    def __init__(self, model: dict, train_cfg=None, test_cfg=None, init_cfg=None):
        # 默认使用teacher模型作为最佳模型
        if test_cfg is None:
            test_cfg = dict()
        test_cfg.update(
            {
                'inference_on': 'student'
            }
        )
        super(ReviewKD, self).__init__(
            dict(
                teacher=build_model(model.teacher), 
                student=build_model(model.student)
            ),
            train_cfg=train_cfg,
            test_cfg=test_cfg,
        )
        self.freeze("teacher")

        self.student_cfg= train_cfg.get('student')
        self.teacher_cfg = train_cfg.get('teacher')

        self.student_in_channels = self.student_cfg.get('channels', [96, 128, 160])      # KeyNetF的后三个stage channels
        self.student_shapes = self.student_cfg.get('shapes', [16,8,4])[::-1]                # KetNetF的后三个stage shape

        # resnet50
        self.teacher_out_channels = self.teacher_cfg.get('channels', [512, 1024, 2048]) # resnet50的后三个stage channels
        self.teacher_shapes = self.teacher_cfg.get('shapes', [16,8,4])[::-1]                # resnet50的后三个stage shape

        # align module
        self.align_module = train_cfg.get('align', 'backbone')

        # 蒸馏损失
        self.distill_loss = build_loss(dict(type='HCL'))

        self.kd_loss_weight = train_cfg.get('kd_loss_weight', 1.0)
        self.kd_warm_up = train_cfg.get('kd_warm_up', 20.0)

        abfs = nn.ModuleList()
        mid_channel = min(512, self.student_in_channels[-1])
        for idx, in_channel in enumerate(self.student_in_channels):
            abfs.append(ABF(in_channel, mid_channel, self.teacher_out_channels[idx], idx < len(self.student_in_channels)-1))
        self.abfs = abfs[::-1]

    def forward_train(self, image, image_meta, **kwargs):
        align_module_proxy = getattr(getattr(self.student, self.align_module), 'forward', None)
        out = {}
        def align_student_module_func(x):
            x = align_module_proxy(x)
            out.update({
                'feature': x
            })
            return x
        # 设置新的forward函数
        setattr(getattr(self.student, self.align_module), 'forward', align_student_module_func)
        student_loss = self.student.forward_train(image, image_meta, **kwargs)
        # 恢复原来forward函数
        setattr(getattr(self.student, self.align_module), 'forward', align_module_proxy)        

        x = out['feature'][::-1]
        s_features = []
        out_features, res_features = self.abfs[0](x[0], out_shape=self.teacher_shapes[0])
        s_features.append(out_features)
        for features, abf, shape, out_shape in zip(x[1:], self.abfs[1:], self.student_shapes[1:], self.teacher_shapes[1:]):
            out_features, res_features = abf(features, res_features, shape, out_shape)
            s_features.insert(0, out_features)

        # 仅计算特征，忽略head计算
        align_module_proxy = getattr(getattr(self.teacher, self.align_module), 'forward', None)
        out = {}
        def align_teacher_module_func(x):
            x = align_module_proxy(x)
            out.update({
                'feature': x
            })
            return x
        # 设置新的forward函数
        setattr(getattr(self.teacher, self.align_module), 'forward', align_teacher_module_func)
        self.teacher.forward_test(image, image_meta)
        # 恢复原来forward函数
        setattr(getattr(self.teacher, self.align_module), 'forward', align_module_proxy)   
        t_features = out['feature']

        # 计算蒸馏损失
        feature_kd_loss = self.distill_loss(s_features, t_features)

        # 任务损失
        losses = student_loss
        # 蒸馏损失
        losses['review_kd_loss'] = feature_kd_loss * min(1, kwargs.get('epoch')/self.kd_warm_up) * self.kd_loss_weight
        return losses

    def simple_test(self, image, image_meta, rescale=True, **kwargs):
        return self.student.simple_test(image, image_meta, rescale, **kwargs)



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
import torch
import torch.nn as nn
import torch.nn.functional as F
from ...builder import DISTILL_LOSSES



# '''
# NST with Polynomial Kernel, where d=2 and c=0
# It can be treated as matching the Gram matrix of two vectorized feature map.
# '''
# class NST(nn.Module):
# 	def __init__(self):
# 		super(NST, self).__init__()

# 	def forward(self, fm_s, fm_t):
# 		fm_s = fm_s.view(fm_s.size(0), fm_s.size(1), -1)
# 		fm_s = F.normalize(fm_s, dim=2)

# 		fm_t = fm_t.view(fm_t.size(0), fm_t.size(1), -1)
# 		fm_t = F.normalize(fm_t, dim=2)

# 		gram_s = self.gram_matrix(fm_s)
# 		gram_t = self.gram_matrix(fm_t)

# 		loss = F.mse_loss(gram_s, gram_t)

# 		return loss

# 	def gram_matrix(self, fm):
# 		return torch.bmm(fm, fm.transpose(1,2))


'''
NST with Polynomial Kernel, where d=2 and c=0
'''
@DISTILL_LOSSES.register_module()
class NST(nn.Module):
	'''
	Like What You Like: Knowledge Distill via Neuron Selectivity Transfer
	https://arxiv.org/pdf/1707.01219.pdf
	'''
	def __init__(self):
		super(NST, self).__init__()

	def forward(self, fm_s, fm_t):
		fm_s = fm_s.view(fm_s.size(0), fm_s.size(1), -1)
		fm_s = F.normalize(fm_s, dim=2)

		fm_t = fm_t.view(fm_t.size(0), fm_t.size(1), -1)
		fm_t = F.normalize(fm_t, dim=2)

		loss = self.poly_kernel(fm_t, fm_t).mean() \
			 + self.poly_kernel(fm_s, fm_s).mean() \
			 - 2 * self.poly_kernel(fm_s, fm_t).mean()

		return loss

	def poly_kernel(self, fm1, fm2):
		fm1 = fm1.unsqueeze(1)
		fm2 = fm2.unsqueeze(2)
		out = (fm1 * fm2).sum(-1).pow(2)

		return out


import sys
sys.path.append('/root/workspace/antgo')
import os
import json
import logging
from antgo.utils import args
from antgo.dataflow.dataset import Cifar10
from antgo.framework.helper.dataset.tfdataset import *

class __CifarDataGenerator(object):
    def __init__(self, flag) -> None:
        self.dataset = Cifar10(flag, './dataset/')
        
        pass
    def __len__(self):
        return self.dataset.size
    
    def __iter__(self):
        for i in range(self.dataset.size):
            sample = self.dataset.sample(i)
            sample.update({
                'tag': f'{i}',
                'image_file': f'{i}.png'
            })
            yield sample
    
def main():
    tgt_folder = './dataset/tfrecord'
    if not os.path.exists(tgt_folder):
        os.makedirs(tgt_folder)
    tfw = TFDataWriter('cifar_10_train', tgt_folder, 100000)
    tfw.write(__CifarDataGenerator('train'))

    tfw = TFDataWriter('cifar_10_test', tgt_folder, 100000)
    tfw.write(__CifarDataGenerator('test'))


if __name__ == "__main__":
    main()


# -*- coding: UTF-8 -*-
# @Time    : 2020/10/26 10:11 上午
# @File    : watch.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.ant.base import *
from antgo.ant.base import _pick_idle_port
from antgo.task.task import *
from antgo.dataflow.common import *
from antgo.dataflow.basic import *
from antgo.dataflow.recorder import *
# from antgo.crowdsource.watch_server import *
from antvis.client.httprpc import *
import json
from antgo.dataflow.dataset.spider_dataset import *
try:
    import queue
except:
    import Queue as queue


class WatchDataRecorder(object):
  def __init__(self, maxsize=30):
    self.queue = queue.Queue()  # 不设置队列最大缓冲
    self.dump_dir = ''
    self.dataset_size = 0
    self.sample_index = 0
    self.tag_dir = ''

  def _transfer_image(self, data):
    try:
      if len(data.shape) == 2:
        if data.dtype == np.uint8:
          transfer_result = data
        else:
          data_min = np.min(data)
          data_max = np.max(data)
          transfer_result = ((data - data_min) / (data_max - data_min) * 255).astype(np.uint8)
      else:
        assert (data.shape[2] == 3)
        transfer_result = data.astype(np.uint8)

      # save path
      if not os.path.exists(self.dump_dir):
        os.makedirs(self.dump_dir)

      file_name = '%s.png' % str(uuid.uuid4())
      image_path = os.path.join(self.dump_dir, file_name)
      imwrite(image_path, transfer_result)
      return file_name
    except:
      logger.error('couldnt transfer image data')
      return None

  def record(self, val):
    # 加入到队列中
    assert(type(val) == dict)

    # 重新组织数据格式
    data = {}
    for key, value in val.items():
      xxyy = key.split('_')
      data_name = xxyy[0]
      if data_name not in data:
        data[data_name] = {}

      # TYPE, TAG
      if(len(xxyy) == 1):
        data[data_name]['data'] = value
        data[data_name]['title'] = data_name
      else:
        if xxyy[-1] == "TYPE":
          data[data_name]['type'] = value
        elif xxyy[-1] == "TAG":
          data[data_name]['tag'] = value
        elif xxyy[-1] == 'ID':
          data[data_name]['id'] = value

    # 转换数据到适合web
    web_data = []
    for name, body in data.items():
      if 'type' in body:
        if body['type'] == 'IMAGE':
          body['width'] = body['data'].shape[1]
          body['height'] = body['data'].shape[0]
          body['data'] = '/static/data/%s'%self._transfer_image(body['data'])

      if 'type' not in body:
        body['type'] = 'STRING'

      if 'tag' not in body:
        body['tag'] = []

      body['dataset_size'] = self.dataset_size
      web_data.append(body)

    # # 加入队列，如果队列满，将阻塞
    # self.queue.put(web_data)
    # rpc 调用

    # 增加样本索引编号
    self.sample_index += 1


class AntWatch(AntBase):
  def __init__(self,
               ant_context,
               ant_name,
               ant_host_ip,
               ant_host_port,
               token,
               data_factory,
               ant_dump_dir,
               ant_task_config, **kwargs):
    super(AntWatch, self).__init__(ant_name, ant_context, token)
    self.ant_data_source = data_factory
    self.ant_dump_dir = ant_dump_dir
    self.ant_context.ant = self
    self.ant_task_config = ant_task_config
    self.context.devices = [int(d) for d in kwargs.get('devices', '').split(',') if d != '']
    self.restore_experiment = kwargs.get('restore_experiment', None)
    self.host_ip = ant_host_ip
    self.host_port = ant_host_port
    self.rpc = None

  def start(self):
    # 1.step 加载挑战任务
    running_ant_task = None
    if self.token is not None:
      # 1.1.step 从平台获取挑战任务配置信息
      response = mlogger.info.challenge.get(command=type(self).__name__)
      if response['status'] == 'ERROR':
        logger.error('couldnt load challenge task')
        self.token = None
      elif response['status'] == 'SUSPEND':
        # prohibit submit challenge task frequently
        # submit only one in one week
        logger.error('prohibit submit challenge task frequently')
        exit(-1)
      elif response['status'] == 'OK':
        content = response['content']

        if 'task' in content:
          challenge_task = create_task_from_json(content)
          if challenge_task is None:
            logger.error('couldnt load challenge task')
            exit(-1)
          running_ant_task = challenge_task
      else:
        # unknow error
        logger.error('unknow error')
        exit(-1)

    if running_ant_task is None:
      # 1.2.step 加载自定义任务配置信息
      custom_task = create_task_from_xml(self.ant_task_config, self.context)
      if custom_task is None:
        logger.error('couldnt load custom task')
        exit(-1)
      running_ant_task = custom_task

    assert (running_ant_task is not None)

    # 2.step 获得实验ID
    experiment_uuid = self.context.experiment_uuid
    if self.restore_experiment is not None:
      experiment_uuid = self.restore_experiment

    # 3.step make experiment folder
    logger.info('build experiment folder')
    if not os.path.exists(os.path.join(self.ant_dump_dir, experiment_uuid)):
      os.makedirs(os.path.join(self.ant_dump_dir, experiment_uuid))
    experiment_dump_dir = os.path.join(self.ant_dump_dir, experiment_uuid)

    # 4.step 配置web服务基本信息
    # 选择端口
    self.host_port = _pick_idle_port(self.host_port)

    # 配置调用
    self.rpc = HttpRpc("v1", "watch-api", "127.0.0.1", self.host_port)

    # 准备素材资源
    static_folder = '/'.join(os.path.dirname(__file__).split('/')[0:-1])
    batch_static_dir = os.path.join(experiment_dump_dir, 'watch')
    if not os.path.exists(batch_static_dir):
      shutil.copytree(os.path.join(static_folder, 'resource', 'watch'), batch_static_dir)

    # 5.step 配置运行
    self.context.recorder = WatchDataRecorder()
    def _run_watch_process():
      # 5.1.step prepare ablation blocks
      logger.info('prepare model ablation blocks')
      ablation_blocks = getattr(self.ant_context.params, 'ablation', [])
      if ablation_blocks is None:
        ablation_blocks = []
      for b in ablation_blocks:
        self.ant_context.deactivate_block(b)

      # 5.2.step 创建数据源
      logger.info('build spider data source')
      ant_watch_dataset = SpiderDataset(queue)
      data_annotation_branch = DataAnnotationBranch(Node.inputs(ant_watch_dataset))

      # 5.3.step 启动推断过程
      logger.info('launch predict process')
      output_dir = os.path.join(experiment_dump_dir, 'batch', 'static', 'data', 'dump')
      if not os.path.exists(output_dir):
        os.makedirs(output_dir)

      with safe_recorder_manager(self.context.recorder):
        self.context.call_infer_process(data_annotation_branch.output(0), dump_dir=output_dir)

    process = threading.Thread(target=_run_watch_process)
    process.daemon = True
    process.start()

    # 6.step 启动bach server
    watch_server_start(experiment_dump_dir, self.host_port, queue)


import math

import numpy as np
import torch
from torch.utils.data import Sampler
import copy
from antgo.framework.helper.runner import get_dist_info


# 将 index iterable 转化为 batch index iterable
# 如 chunk([4, 2, 3, 1], 2) ==> [[4, 2], [3, 1]]
def _chunk(iterable, chunk_size):
    ret = []
    for record in iterable:
        if record is None:
            # 跳过 None，会造成组成的batch中，不满足预先指定的方案
            continue

        ret.append(record)
        if len(ret) == chunk_size:
            yield ret
            ret = []
    if ret:
        yield ret


class KVSampler(Sampler):
    def __init__(self, dataset, samples_per_gpu=1, shuffle=False, drop_last=False, strategy=None) -> None:
        super().__init__(dataset)
        assert hasattr(dataset, 'flag')
        self.flag = dataset.flag.astype(np.int64)
        # strtegy 中保存样本数据源比例 ({0: 2, 1: 1} 表示flag=0和flag=1在一个batch中的个数)
        self.strategy = strategy

        self.shuffle = shuffle
        self.samples_per_gpu = samples_per_gpu
        self.drop_last = drop_last
        self.dataset = dataset

        self.label_and_unlabel_group_sizes = np.bincount(self.flag)
        self.num_samples = 0
        for i, size in enumerate(self.label_and_unlabel_group_sizes):
            self.num_samples += int(np.ceil(
                size / self.samples_per_gpu)) * self.samples_per_gpu

        self.source_indices = []
        self.source_num_per_batch = []
        self.source_key_per_batch = []
        if self.strategy is not None:
            self.source_indices = [None for _ in range(len(self.strategy))]
            self.source_num_per_batch = [None for _ in range(len(self.strategy))]
            self.source_key_per_batch = [None for _ in range(len(self.strategy))]    
            for k,v in self.strategy.items():
                k = int(k)
                self.source_indices[k] = np.where(self.flag == int(k))[0]
                self.source_num_per_batch[k] = int(v)
                self.source_key_per_batch[k] = int(k)
            
            assert(np.sum(self.source_num_per_batch) == samples_per_gpu)

    def sampling_by_normal(self):
        # 每次返回一个batch_size
        indices = list(range(len(self.dataset)))
        if self.shuffle:
            np.random.shuffle(indices)

        if not self.drop_last:
            # add extra samples to make it evenly divisible
            padding_size = self.num_samples - len(indices)
            if padding_size <= len(indices):
                indices += indices[:padding_size]
            else:
                indices += (indices * math.ceil(padding_size / len(indices)))[:padding_size]
        else:
            # remove tail of data to make it evenly divisible.
            indices = indices[:self.num_samples]

        return _chunk(iter(indices), self.samples_per_gpu)        

    def sampling_by_strategy(self):
        source_indices = copy.deepcopy(self.source_indices)
        if self.shuffle:
            for index in range(len(source_indices)):
                np.random.shuffle(source_indices[index])

        offset_in_source = [0 for _ in range(len(source_indices))]
        selected_indices = []
        for batch_i in range(self.num_samples // self.samples_per_gpu):
            # 逐数据源挑选数据
            selected_source_indices = [None for _ in range(len(source_indices))]
            for source_i in range(len(source_indices)):
                offset_in_source_i = offset_in_source[source_i]
                if offset_in_source_i + self.source_num_per_batch[source_i] <= len(source_indices[source_i]):
                    selected_source_indices[source_i] = source_indices[source_i][offset_in_source_i:offset_in_source_i+self.source_num_per_batch[source_i]]
                    offset_in_source[source_i] += self.source_num_per_batch[source_i]
                else:
                    num_extra = offset_in_source_i + self.source_num_per_batch[source_i] - len(source_indices[source_i])
                    source_indices_extra = source_indices[source_i][offset_in_source_i:]

                    if len(source_indices_extra) != 0:
                        selected_source_indices[source_i] = np.concatenate(
                                [source_indices_extra, np.random.choice(source_indices_extra, num_extra)])
                    else:
                        selected_source_indices[source_i] = np.random.choice(source_indices[source_i], num_extra)
                    
                    offset_in_source[source_i] = len(source_indices[source_i])

            # 合并
            selected_indices_in_batch = np.concatenate(selected_source_indices)
            selected_indices_in_batch = selected_indices_in_batch.astype(np.int64).tolist()

            # 扩展
            selected_indices.extend(selected_indices_in_batch)

        return _chunk(iter(selected_indices), self.samples_per_gpu)     

    def __iter__(self):
        if self.strategy is None:
            return self.sampling_by_normal()
        else:
            return self.sampling_by_strategy()

    def __len__(self):
        return self.num_samples // self.samples_per_gpu


class DistributedKVSampler(torch.utils.data.distributed.DistributedSampler):
    def __init__(self, 
                 dataset, 
                 samples_per_gpu=1, 
                 num_replicas=1,
                 rank=None, 
                 shuffle=True,
                 drop_last=False,
                 seed=0, strategy=None) -> None:
        super().__init__(dataset, num_replicas= num_replicas, rank=rank, shuffle=shuffle, drop_last=drop_last, seed=seed)        
        self.batch_size = samples_per_gpu
        _rank, _num_replicas = get_dist_info()
        if num_replicas is None:
            num_replicas = _num_replicas
        if rank is None:
            rank = _rank
        self.dataset = dataset
        self.samples_per_gpu = samples_per_gpu
        self.num_replicas = num_replicas

        self.shuffle = shuffle
        self.drop_last = drop_last

        assert hasattr(self.dataset, 'flag')
        self.flag = self.dataset.flag.astype(np.int64)
        self.group_sizes = np.bincount(self.flag)

        self.num_samples = 0
        for i, j in enumerate(self.group_sizes):
            self.num_samples += int(
                math.ceil(self.group_sizes[i] * 1.0 / self.samples_per_gpu /
                          self.num_replicas)) * self.samples_per_gpu
        self.total_size = self.num_samples * self.num_replicas

        # strtegy 中保存样本数据源比例 ({0: 2, 1: 1} 表示flag=0和flag=1在一个batch中的个数)
        self.strategy = strategy
        self.source_indices = []
        self.source_num_per_batch = []
        self.source_key_per_batch = []
        if self.strategy is not None:
            self.source_indices = [None for _ in range(len(self.strategy))]
            self.source_num_per_batch = [None for _ in range(len(self.strategy))]
            self.source_key_per_batch = [None for _ in range(len(self.strategy))]               
            for k,v in self.strategy.items():
                k = int(k)

                self.source_indices[k] = np.where(self.flag == int(k))[0]
                self.source_num_per_batch[k] = int(v)
                self.source_key_per_batch[k] = int(k)                
            
            assert(np.sum(self.source_num_per_batch) == samples_per_gpu)        

    def sampling_by_normal(self):
        # 每次返回一个batch_size
        g = torch.Generator()
        g.manual_seed(self.epoch + self.seed)

        indices = []
        for i, size in enumerate(self.group_sizes):
            if size > 0:
                indice = np.where(self.flag == i)[0]
                assert len(indice) == size
                indice = indice[list(
                    torch.randperm(int(size), generator=g).numpy())].tolist()
                
                extra = int(
                    math.ceil(
                        size * 1.0 / self.samples_per_gpu / self.num_replicas)
                ) * self.samples_per_gpu * self.num_replicas - len(indice)

                # pad indice
                tmp = indice.copy()
                for _ in range(extra // size):
                    indice.extend(tmp)
                indice.extend(tmp[:extra % size])
                indices.extend(indice)
        assert len(indices) == self.total_size

        indices = [
            indices[j] for i in list(
                torch.randperm(
                    len(indices) // self.samples_per_gpu, generator=g))
            for j in range(i * self.samples_per_gpu, (i + 1) *
                           self.samples_per_gpu)
        ]

        # subsample
        offset = self.num_samples * self.rank
        indices = indices[offset:offset + self.num_samples]
        assert len(indices) == self.num_samples
        return _chunk(iter(indices), self.samples_per_gpu)

    def sampling_by_strategy(self):
        source_indices = copy.deepcopy(self.source_indices)

        offset_in_source = [0 for _ in range(len(source_indices))]
        selected_indices = []
        perfect_selected_source_indices = [[] for _ in range(len(source_indices))]
        for batch_i in range(self.total_size // self.samples_per_gpu):
            # 扩展数据满足个数要求（每张卡需要有相同 个数样本）
            selected_source_indices = [None for _ in range(len(source_indices))]
            for source_i in range(len(source_indices)):
                offset_in_source_i = offset_in_source[source_i]
                if offset_in_source_i + self.source_num_per_batch[source_i] <= len(source_indices[source_i]):
                    selected_source_indices[source_i] = source_indices[source_i][offset_in_source_i:offset_in_source_i+self.source_num_per_batch[source_i]]
                    offset_in_source[source_i] += self.source_num_per_batch[source_i]
                else:
                    num_extra = offset_in_source_i + self.source_num_per_batch[source_i] - len(source_indices[source_i])
                    source_indices_extra = source_indices[source_i][offset_in_source_i:]

                    if len(source_indices_extra) != 0:
                        selected_source_indices[source_i] = np.concatenate(
                                [source_indices_extra, np.random.choice(source_indices_extra, num_extra)])
                    else:
                        selected_source_indices[source_i] = np.random.choice(source_indices[source_i], num_extra)
                    
                    offset_in_source[source_i] = len(source_indices[source_i])

            # 填充到candidate_selected_source_indices中的索引已经满足
            # 每张卡都有相同batch_size的个数
            for source_i in range(len(source_indices)):
                perfect_selected_source_indices[source_i].extend(selected_source_indices[source_i].astype(np.int64).tolist())

        # step1: 先获取分配给每张卡内的固定数据索引
        # step2: 在固定索引内部进行shuffle
        shuffle_selected_source_indices = [None for _ in range(len(source_indices))]
        for source_i in range(len(source_indices)):
            k = self.source_key_per_batch[source_i]
            assert(k == source_i)
            v = self.source_num_per_batch[source_i]
            start_index = self.rank * (self.num_samples // self.samples_per_gpu * v)
            end_index = (self.rank+1) * (self.num_samples // self.samples_per_gpu * v)
            shuffle_selected_source_indices[source_i] = perfect_selected_source_indices[source_i][start_index:end_index]
            np.random.shuffle(shuffle_selected_source_indices[source_i])

        # 合并
        selected_indices = []
        for batch_i in range(self.num_samples // self.samples_per_gpu):
            for source_i in range(len(shuffle_selected_source_indices)):
                v = self.source_num_per_batch[source_i]
                selected_indices.extend(shuffle_selected_source_indices[source_i][batch_i*v: (batch_i+1)*v])

        assert(len(selected_indices) % self.samples_per_gpu == 0)
        return _chunk(iter(selected_indices), self.samples_per_gpu)     

    def __iter__(self):
        if self.strategy is None:
            return self.sampling_by_normal()
        else:
            return self.sampling_by_strategy()

    def __len__(self):
        return self.num_samples

    def set_epoch(self, epoch):
        self.epoch = epoch



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
import torch
import torch.nn as nn
import torch.nn.functional as F
from ...builder import DISTILL_LOSSES


@DISTILL_LOSSES.register_module()
class AB(nn.Module):
	'''
	Knowledge Transfer via Distillation of Activation Boundaries Formed by Hidden Neurons
	https://arxiv.org/pdf/1811.03233.pdf
	'''
	def __init__(self, margin):
		super(AB, self).__init__()

		self.margin = margin

	def forward(self, fm_s, fm_t):
		# fm befor activation
		loss = ((fm_s + self.margin).pow(2) * ((fm_s > -self.margin) & (fm_t <= 0)).float() +
			    (fm_s - self.margin).pow(2) * ((fm_s <= self.margin) & (fm_t > 0)).float())
		loss = loss.mean()

		return loss


import bisect
import collections
import copy
import math
from collections import defaultdict
from xml.etree.ElementTree import iselement

import numpy as np
from antgo.framework.helper.utils import build_from_cfg, print_log
import torch
from torch.utils.data.dataset import ConcatDataset as _ConcatDataset
from .builder import DATASETS, PIPELINES
import json


@DATASETS.register_module()
class IterableDatasetFilter(torch.utils.data.IterableDataset):
    def __init__(self, dataset, not_in_anns=None) -> None:
        self.dataset = dataset
        self.not_in_map = {}
        for ann_file in not_in_anns:
            with open(ann_file, 'r') as fp:
                content = json.load(fp)
            for sample_info in content:
                sample_id = f'{sample_info["image_file"].split("/")[1]}'
                if sample_id.endswith('.png') or sample_id.endswith('.jpg'):
                    sample_id = sample_id[:-4]
                elif sample_id.endswith('.jpeg'):
                    sample_id = sample_id[:-5]
                self.not_in_map[sample_id] = True

    def __iter__(self):
        dataset_iter = iter(self.dataset)
        while True:
            try:
                sample = next(dataset_iter)
                if len(self.not_in_map) > 0:
                    sample_tag = sample["image_meta"]["tag"]                   
                    sample_id = f"{sample_tag}_{sample['image_meta']['image_file'].replace('/','-')}"
                    if sample_id in self.not_in_map:
                        continue
                    
                yield sample
            except StopIteration:
                break

    def __len__(self):
        # 这里返回的是最大数，实际数量可能少于此（部分样本会被过滤）
        return len(self.dataset)



# -*- coding: UTF-8 -*-
# @Time    : 2022/9/13 16:09
# @File    : __init__.py.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function



# encoding=utf-8
# @Time    : 17-5-3
# @File    : base.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.task.task import *


class AntMeasure(object):
  _BASE_MEASURE=True

  def __init__(self, task,name=None):
    self.task = task
    if name != None:
      self.name = name
    else:
      self.name = AntMeasure.__class__.__name__

    self._is_inverse = False
    self._is_support_rank = False
    self._support_rank_index = 0
    self._crowdsource = False
    self.larger_is_better = True

  @property
  def is_support_rank(self):
    return self._is_support_rank
  @is_support_rank.setter
  def is_support_rank(self,val):
    self._is_support_rank = val

  @property
  def support_rank_index(self):
    return self._support_rank_index
  @support_rank_index.setter
  def support_rank_index(self, index):
    self._support_rank_index = index

  @property
  def crowdsource(self):
    return self._crowdsource
  @crowdsource.setter
  def crowdsource(self, val):
    self._crowdsource = val


# Copyright (c) OpenMMLab. All rights reserved.
import itertools

import numpy as np
import torch
from antgo.framework.helper.runner import get_dist_info
from torch.utils.data.sampler import Sampler

from antgo.framework.helper.utils.dist_utils import sync_random_seed


class InfiniteGroupBatchSampler(Sampler):
    """Similar to `BatchSampler` warping a `GroupSampler. It is designed for
    iteration-based runners like `IterBasedRunner` and yields a mini-batch
    indices each time, all indices in a batch should be in the same group.

    The implementation logic is referred to
    https://github.com/facebookresearch/detectron2/blob/main/detectron2/data/samplers/grouped_batch_sampler.py

    Args:
        dataset (object): The dataset.
        batch_size (int): When model is :obj:`DistributedDataParallel`,
            it is the number of training samples on each GPU.
            When model is :obj:`DataParallel`, it is
            `num_gpus * samples_per_gpu`.
            Default : 1.
        world_size (int, optional): Number of processes participating in
            distributed training. Default: None.
        rank (int, optional): Rank of current process. Default: None.
        seed (int): Random seed. Default: 0.
        shuffle (bool): Whether shuffle the indices of a dummy `epoch`, it
            should be noted that `shuffle` can not guarantee that you can
            generate sequential indices because it need to ensure
            that all indices in a batch is in a group. Default: True.
    """  # noqa: W605

    def __init__(self,
                 dataset,
                 batch_size=1,
                 world_size=None,
                 rank=None,
                 seed=0,
                 shuffle=True):
        _rank, _world_size = get_dist_info()
        if world_size is None:
            world_size = _world_size
        if rank is None:
            rank = _rank
        self.rank = rank
        self.world_size = world_size
        self.dataset = dataset
        self.batch_size = batch_size
        # In distributed sampling, different ranks should sample
        # non-overlapped data in the dataset. Therefore, this function
        # is used to make sure that each rank shuffles the data indices
        # in the same order based on the same seed. Then different ranks
        # could use different indices to select non-overlapped data from the
        # same data list.
        self.seed = sync_random_seed(seed)
        self.shuffle = shuffle

        assert hasattr(self.dataset, 'flag')
        self.flag = self.dataset.flag
        self.group_sizes = np.bincount(self.flag)
        # buffer used to save indices of each group
        self.buffer_per_group = {k: [] for k in range(len(self.group_sizes))}

        self.size = len(dataset)
        self.indices = self._indices_of_rank()

    def _infinite_indices(self):
        """Infinitely yield a sequence of indices."""
        g = torch.Generator()
        g.manual_seed(self.seed)
        while True:
            if self.shuffle:
                yield from torch.randperm(self.size, generator=g).tolist()

            else:
                yield from torch.arange(self.size).tolist()

    def _indices_of_rank(self):
        """Slice the infinite indices by rank."""
        yield from itertools.islice(self._infinite_indices(), self.rank, None,
                                    self.world_size)

    def __iter__(self):
        # once batch size is reached, yield the indices
        for idx in self.indices:
            flag = self.flag[idx]
            group_buffer = self.buffer_per_group[flag]
            group_buffer.append(idx)
            if len(group_buffer) == self.batch_size:
                yield group_buffer[:]
                del group_buffer[:]

    def __len__(self):
        """Length of base dataset."""
        return self.size

    def set_epoch(self, epoch):
        """Not supported in `IterationBased` runner."""
        raise NotImplementedError


class InfiniteBatchSampler(Sampler):
    """Similar to `BatchSampler` warping a `DistributedSampler. It is designed
    iteration-based runners like `IterBasedRunner` and yields a mini-batch
    indices each time.

    The implementation logic is referred to
    https://github.com/facebookresearch/detectron2/blob/main/detectron2/data/samplers/grouped_batch_sampler.py

    Args:
        dataset (object): The dataset.
        batch_size (int): When model is :obj:`DistributedDataParallel`,
            it is the number of training samples on each GPU,
            When model is :obj:`DataParallel`, it is
            `num_gpus * samples_per_gpu`.
            Default : 1.
        world_size (int, optional): Number of processes participating in
            distributed training. Default: None.
        rank (int, optional): Rank of current process. Default: None.
        seed (int): Random seed. Default: 0.
        shuffle (bool): Whether shuffle the dataset or not. Default: True.
    """  # noqa: W605

    def __init__(self,
                 dataset,
                 batch_size=1,
                 world_size=None,
                 rank=None,
                 seed=0,
                 shuffle=True):
        _rank, _world_size = get_dist_info()
        if world_size is None:
            world_size = _world_size
        if rank is None:
            rank = _rank
        self.rank = rank
        self.world_size = world_size
        self.dataset = dataset
        self.batch_size = batch_size
        # In distributed sampling, different ranks should sample
        # non-overlapped data in the dataset. Therefore, this function
        # is used to make sure that each rank shuffles the data indices
        # in the same order based on the same seed. Then different ranks
        # could use different indices to select non-overlapped data from the
        # same data list.
        self.seed = sync_random_seed(seed)
        self.shuffle = shuffle
        self.size = len(dataset)
        self.indices = self._indices_of_rank()

    def _infinite_indices(self):
        """Infinitely yield a sequence of indices."""
        g = torch.Generator()
        g.manual_seed(self.seed)
        while True:
            if self.shuffle:
                yield from torch.randperm(self.size, generator=g).tolist()

            else:
                yield from torch.arange(self.size).tolist()

    def _indices_of_rank(self):
        """Slice the infinite indices by rank."""
        yield from itertools.islice(self._infinite_indices(), self.rank, None,
                                    self.world_size)

    def __iter__(self):
        # once batch size is reached, yield the indices
        batch_buffer = []
        for idx in self.indices:
            batch_buffer.append(idx)
            if len(batch_buffer) == self.batch_size:
                yield batch_buffer
                batch_buffer = []

    def __len__(self):
        """Length of base dataset."""
        return self.size

    def set_epoch(self, epoch):
        """Not supported in `IterationBased` runner."""
        raise NotImplementedError



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
import torch
import torch.nn as nn
import torch.nn.functional as F
from ...builder import DISTILL_LOSSES


@DISTILL_LOSSES.register_module()
class SoftTarget(nn.Module):
	'''
	Distilling the Knowledge in a Neural Network
	https://arxiv.org/pdf/1503.02531.pdf
	'''
	def __init__(self, T):
		super(SoftTarget, self).__init__()
		self.T = T

	def forward(self, out_s, out_t):
		loss = F.kl_div(F.log_softmax(out_s/self.T, dim=1),
						F.softmax(out_t/self.T, dim=1),
						reduction='batchmean') * self.T * self.T

		return loss


from typing import Dict, Sequence

import torch.nn as nn
import torch.nn.functional as F
from antgo.framework.helper.cnn import build_activation_layer, build_norm_layer
from antgo.framework.helper.runner import BaseModule, ModuleList
from .cls_head import ClsHead
from ...builder import *

class LinearBlock(nn.Module):

    def __init__(self,
                 in_channels,
                 out_channels,
                 dropout_rate=0.,
                 norm_cfg=None,
                 act_cfg=None,
                 init_cfg=None):
        super().__init__()
        self.fc = nn.Linear(in_channels, out_channels)

        self.norm = None
        self.act = None
        self.dropout = None

        if norm_cfg is not None:
            self.norm = build_norm_layer(norm_cfg, out_channels)[1]
        if act_cfg is not None:
            self.act = build_activation_layer(act_cfg)
        if dropout_rate > 0:
            self.dropout = nn.Dropout(p=dropout_rate)

    def forward(self, x):
        x = self.fc(x)
        if self.norm is not None:
            x = self.norm(x)
        if self.act is not None:
            x = self.act(x)
        if self.dropout is not None:
            x = self.dropout(x)
        return x


@HEADS.register_module()
class StackedLinearClsHead(ClsHead):
    """Classifier head with several hidden fc layer and a output fc layer.

    Args:
        num_classes (int): Number of categories.
        in_channels (int): Number of channels in the input feature map.
        mid_channels (Sequence): Number of channels in the hidden fc layers.
        dropout_rate (float): Dropout rate after each hidden fc layer,
            except the last layer. Defaults to 0.
        norm_cfg (dict, optional): Config dict of normalization layer after
            each hidden fc layer, except the last layer. Defaults to None.
        act_cfg (dict, optional): Config dict of activation function after each
            hidden layer, except the last layer. Defaults to use "ReLU".
    """

    def __init__(self,
                 num_classes,
                 in_channels,
                 mid_channels,
                 dropout_rate = 0.,
                 norm_cfg = None,
                 act_cfg = dict(type='ReLU'),
                 **kwargs):
        super(StackedLinearClsHead, self).__init__(**kwargs)
        assert num_classes > 0, \
            f'`num_classes` of StackedLinearClsHead must be a positive ' \
            f'integer, got {num_classes} instead.'
        self.num_classes = num_classes

        self.in_channels = in_channels

        assert isinstance(mid_channels, Sequence), \
            f'`mid_channels` of StackedLinearClsHead should be a sequence, ' \
            f'instead of {type(mid_channels)}'
        self.mid_channels = mid_channels

        self.dropout_rate = dropout_rate
        self.norm_cfg = norm_cfg
        self.act_cfg = act_cfg

        self._init_layers()

    def _init_layers(self):
        self.layers = ModuleList()
        in_channels = self.in_channels
        for hidden_channels in self.mid_channels:
            self.layers.append(
                LinearBlock(
                    in_channels,
                    hidden_channels,
                    dropout_rate=self.dropout_rate,
                    norm_cfg=self.norm_cfg,
                    act_cfg=self.act_cfg))
            in_channels = hidden_channels

        self.layers.append(
            LinearBlock(
                self.mid_channels[-1],
                self.num_classes,
                dropout_rate=0.,
                norm_cfg=None,
                act_cfg=None))

    def init_weights(self):
        self.layers.init_weights()

    def pre_logits(self, x):
        if isinstance(x, tuple):
            x = x[-1]
        for layer in self.layers[:-1]:
            x = layer(x)
        return x

    @property
    def fc(self):
        return self.layers[-1]

    def forward_train(self, x, gt_label, **kwargs):
        x = self.pre_logits(x)
        cls_score = self.fc(x)
        losses = self.loss(cls_score, gt_label, **kwargs)
        return losses

    def simple_test(self, x, softmax=True, post_process=True):
        """Inference without augmentation.

        Args:
            x (tuple[Tensor]): The input features.
                Multi-stage inputs are acceptable but only the last stage will
                be used to classify. The shape of every item should be
                ``(num_samples, in_channels)``.
            softmax (bool): Whether to softmax the classification score.
            post_process (bool): Whether to do post processing the
                inference results. It will convert the output to a list.

        Returns:
            Tensor | list: The inference results.

                - If no post processing, the output is a tensor with shape
                  ``(num_samples, num_classes)``.
                - If post processing, the output is a multi-dimentional list of
                  float and the dimensions are ``(num_samples, num_classes)``.
        """
        x = self.pre_logits(x)
        cls_score = self.fc(x)

        if softmax:
            pred = (
                F.softmax(cls_score, dim=1) if cls_score is not None else None)
        else:
            pred = cls_score

        return pred



import logging
import os
import shutil
import yaml 
import json
import subprocess
from antgo import config
from antgo.script.base import *


# 提交任务运行 (仅作为debug时使用，在本地机器上运行任务)
def local_submit_process_func(project_name, sys_argv, gpu_num, cpu_num, memory_size, task_name=None):   
    # 前提假设，调用此函数前当前目录下需要存在项目代码
    with open(os.path.join(config.AntConfig.task_factory,f'{project_name}.json'), 'r') as fp:
        project_info = json.load(fp)

    if task_name is not None:
        extra_config = prepare_extra_config(task_name, project_info)
        if extra_config is None:
            return False

        with open('./extra-config.json', 'w') as fp:
            json.dump(extra_config, fp)      
        sys_argv += " --extra-config=./extra-config.json"

    # 后台运行
    process = subprocess.Popen(sys_argv, shell=True)
    with open(os.path.join(config.AntConfig.factory, '.local_submit_process.pid'), 'w') as fp:
        fp.write(f'{process.pid}')

    # 删除临时配置：
    if os.path.exists('./extra-config.py'):
        os.remove('./extra-config.py')
    return True


# 检查任务资源是否满足
def local_submit_resource_check_func(gpu_num, cpu_num, memory_size):
    # # 单位时间内仅可以一个任务在执行
    # if not os.path.exists(os.path.join(config.AntConfig.factory, '.local_submit_process.pid')):
    #     return True
    
    # with open(os.path.join(config.AntConfig.factory, '.local_submit_process.pid'), 'r') as fp:
    #     process_pid = fp.read().strip()
        
    # # 检查process_pid是否在运行
    # process_dir = os.path.join('/proc', str(process_pid))
    # if os.path.exists(process_dir):
    #     return False
    
    return True